@@ -6,33 +6,34 @@ mod highlighter;
 use ansi_formatter::ansi_format;
 use colored::control::SHOULD_COLORIZE;
 use completer::NumbatCompleter;
-use config::{ColorMode, Config, ExchangeRateFetchingPolicy, IntroBanner, PrettyPrintMode};
+use config::{
+    ColorMode, Config, EditMode, ExchangeRateFetchingPolicy, IntroBanner, PrettyPrintMode,
+};
 use highlighter::NumbatHighlighter;
 
 use itertools::Itertools;
 use numbat::command::{CommandControlFlow, CommandRunner};
-use numbat::compact_str::CompactString;
-use numbat::diagnostic::ErrorDiagnostic;
-use numbat::markup as m;
+use numbat::diagnostic::{ErrorDiagnostic, ResolverDiagnostic};
 use numbat::module_importer::{BuiltinModuleImporter, ChainedImporter, FileSystemImporter};
 use numbat::pretty_print::PrettyPrint;
 use numbat::resolver::CodeSource;
 use numbat::session_history::{ParseEvaluationResult, SessionHistory};
 use numbat::{Context, NumbatError};
 use numbat::{InterpreterSettings, NameResolutionError};
+use numbat::{RuntimeErrorKind, markup as m};
 
-use anyhow::{bail, Context as AnyhowContext, Result};
+use anyhow::{Context as AnyhowContext, Result, bail};
 use clap::Parser;
 use rustyline::config::Configurer;
 use rustyline::{
-    error::ReadlineError, history::DefaultHistory, Completer, Editor, Helper, Hinter, Validator,
+    Completer, Editor, Helper, Hinter, Validator, error::ReadlineError, history::DefaultHistory,
 };
 use rustyline::{EventHandler, Highlighter, KeyCode, KeyEvent, Modifiers};
 
-use std::io::IsTerminal;
-use std::path::PathBuf;
+use std::io::{IsTerminal, Write};
+use std::path::{Path, PathBuf};
 use std::sync::{Arc, Mutex};
-use std::{fs, thread};
+use std::{env, fs, thread};
 
 #[derive(Debug, PartialEq, Eq)]
 pub enum ExitStatus {
@@ -107,7 +108,7 @@ enum ExecutionMode {
 }
 
 impl ExecutionMode {
-    fn exit_status_in_case_of_error(&self) -> ControlFlow {
+    fn exit_status_in_case_of_error(self) -> ControlFlow {
         if matches!(self, ExecutionMode::Normal) {
             ControlFlow::Break(ExitStatus::Error)
         } else {
@@ -300,6 +301,10 @@ impl Cli {
         let history_path = self.get_history_path()?;
 
         let mut rl = Editor::<NumbatHelper, DefaultHistory>::new()?;
+        rl.set_edit_mode(match self.config.edit_mode {
+            EditMode::Emacs => rustyline::EditMode::Emacs,
+            EditMode::Vi => rustyline::EditMode::Vi,
+        });
         rl.set_max_history_size(1000)
             .context("Error while configuring history size")?;
         rl.set_completion_type(rustyline::CompletionType::List);
@@ -309,7 +314,7 @@ impl Cli {
                 modules: self.context.lock().unwrap().list_modules().collect(),
                 all_timezones: jiff::tz::db()
                     .available()
-                    .map(CompactString::from)
+                    .map(|name| name.as_str().into())
                     .collect(),
             },
             highlighter: NumbatHighlighter {
@@ -343,7 +348,7 @@ impl Cli {
             }
         }
 
-        let result = self.repl_loop(&mut rl, interactive);
+        let result = self.repl_loop(&mut rl, interactive, &history_path);
 
         if interactive {
             rl.save_history(&history_path).context(format!(
@@ -359,6 +364,7 @@ impl Cli {
         &mut self,
         rl: &mut Editor<NumbatHelper, DefaultHistory>,
         interactive: bool,
+        history_path: &Path,
     ) -> Result<()> {
         let mut cmd_runner = CommandRunner::<Editor<NumbatHelper, DefaultHistory>>::new()
             .print_with(|m| println!("{}", ansi_format(m, true)))
@@ -367,7 +373,7 @@ impl Cli {
                 Err(_) => CommandControlFlow::Return,
             })
             .enable_save(SessionHistory::default())
-            .enable_reset(Self::make_fresh_context)
+            .enable_reset()
             .enable_quit();
 
         loop {
@@ -379,16 +385,38 @@ impl Cli {
                     }
 
                     rl.add_history_entry(&line)?;
-
                     let mut ctx = self.context.lock().unwrap();
+
+                    if interactive && rl.append_history(history_path).is_err() {
+                        ctx.print_diagnostic(
+                            ResolverDiagnostic {
+                                resolver: ctx.resolver(),
+                                error: &ctx.runtime_error(RuntimeErrorKind::HistoryWrite(
+                                    history_path.to_owned(),
+                                )),
+                            },
+                            colored::control::SHOULD_COLORIZE.should_colorize(),
+                        );
+                    }
+
                     match cmd_runner.try_run_command(&line, &mut ctx, rl) {
                         Ok(cf) => match cf {
                             CommandControlFlow::Continue => continue,
                             CommandControlFlow::Return => return Ok(()),
+                            CommandControlFlow::Reset => {
+                                *ctx = Self::make_fresh_context();
+                                continue;
+                            }
                             CommandControlFlow::NotACommand => {}
                         },
                         Err(err) => {
-                            ctx.print_diagnostic(*err);
+                            ctx.print_diagnostic(
+                                ResolverDiagnostic {
+                                    resolver: ctx.resolver(),
+                                    error: &*err,
+                                },
+                                colored::control::SHOULD_COLORIZE.should_colorize(),
+                            );
                             continue;
                         }
                     }
@@ -520,7 +548,14 @@ impl Cli {
                 execution_mode.exit_status_in_case_of_error()
             }
             Err(NumbatError::RuntimeError(e)) => {
-                self.print_diagnostic(e);
+                let ctx = self.context.lock().unwrap();
+                ctx.print_diagnostic(
+                    ResolverDiagnostic {
+                        resolver: ctx.resolver(),
+                        error: &e,
+                    },
+                    colored::control::SHOULD_COLORIZE.should_colorize(),
+                );
                 execution_mode.exit_status_in_case_of_error()
             }
         };
@@ -532,7 +567,10 @@ impl Cli {
     }
 
     fn print_diagnostic(&mut self, error: impl ErrorDiagnostic) {
-        self.context.lock().unwrap().print_diagnostic(error)
+        self.context
+            .lock()
+            .unwrap()
+            .print_diagnostic(error, colored::control::SHOULD_COLORIZE.should_colorize())
     }
 
     fn get_config_path() -> PathBuf {
@@ -567,6 +605,14 @@ impl Cli {
     }
 
     fn get_history_path(&self) -> Result<PathBuf> {
+        if let Ok(history) = env::var("NUMBAT_HISTORY") {
+            let history_path = PathBuf::from(history);
+            if let Some(parent) = history_path.parent() {
+                fs::create_dir_all(parent).ok();
+            }
+            return Ok(history_path);
+        }
+
         let data_dir = dirs::data_dir()
             .unwrap_or_else(|| PathBuf::from("."))
             .join("numbat");
@@ -600,7 +646,9 @@ fn generate_config() -> Result<()> {
         "A default configuration has been written to '{}'.",
         config_file_path.to_string_lossy()
     );
-    println!("Open the file in a text editor. Modify whatever you want to change and remove the other fields");
+    println!(
+        "Open the file in a text editor. Modify whatever you want to change and remove the other fields"
+    );
 
     Ok(())
 }
@@ -617,7 +665,8 @@ fn main() {
     }
 
     if let Err(e) = Cli::new(args).and_then(|mut cli| cli.run()) {
-        eprintln!("{e:#}");
+        let mut stdout = termcolor::StandardStream::stderr(termcolor::ColorChoice::Never);
+        writeln!(stdout, "{e:#}").unwrap();
         std::process::exit(1);
     }
 }
