@@ -5,15 +5,16 @@ use std::sync::{Arc, Mutex};
 use wasm_bindgen::prelude::*;
 
 use numbat::buffered_writer::BufferedWriter;
-use numbat::diagnostic::ErrorDiagnostic;
-use numbat::help::help_markup;
+use numbat::command::{CommandControlFlow, CommandRunner};
+use numbat::diagnostic::{ErrorDiagnostic, ResolverDiagnostic};
+use numbat::help::basic_help_markup;
 use numbat::html_formatter::{HtmlFormatter, HtmlWriter};
 use numbat::markup::Formatter;
 use numbat::module_importer::BuiltinModuleImporter;
 use numbat::pretty_print::PrettyPrint;
 use numbat::resolver::CodeSource;
-use numbat::{markup as m, NameResolutionError, NumbatError};
 use numbat::{Context, InterpreterSettings};
+use numbat::{NameResolutionError, NumbatError, markup as m};
 
 use jquery_terminal_formatter::{JqueryTerminalFormatter, JqueryTerminalWriter};
 
@@ -51,6 +52,23 @@ impl InterpreterOutput {
     }
 }
 
+#[wasm_bindgen]
+#[derive(Debug, Clone)]
+pub struct CommandResult {
+    pub is_command: bool,
+    pub should_clear: bool,
+    pub should_reset: bool,
+    output: String,
+}
+
+#[wasm_bindgen]
+impl CommandResult {
+    #[wasm_bindgen(getter)]
+    pub fn output(&self) -> String {
+        self.output.clone()
+    }
+}
+
 #[wasm_bindgen]
 impl Numbat {
     pub fn new(load_prelude: bool, enable_pretty_printing: bool, format_type: FormatType) -> Self {
@@ -143,32 +161,72 @@ impl Numbat {
                 | NameResolutionError::ReservedIdentifier(_)),
             )) => self.print_diagnostic(&e),
             Err(NumbatError::TypeCheckError(e)) => self.print_diagnostic(&e),
-            Err(NumbatError::RuntimeError(e)) => self.print_diagnostic(&e),
+            Err(NumbatError::RuntimeError(e)) => self.print_diagnostic(&ResolverDiagnostic {
+                resolver: self.ctx.resolver(),
+                error: &e,
+            }),
         }
     }
 
-    pub fn print_environment(&self) -> JsValue {
-        self.format(&self.ctx.print_environment(), false).into()
-    }
-
-    pub fn print_functions(&self) -> JsValue {
-        self.format(&self.ctx.print_functions(), false).into()
-    }
-
-    pub fn print_dimensions(&self) -> JsValue {
-        self.format(&self.ctx.print_dimensions(), false).into()
-    }
-
-    pub fn print_variables(&self) -> JsValue {
-        self.format(&self.ctx.print_variables(), false).into()
-    }
-
-    pub fn print_units(&self) -> JsValue {
-        self.format(&self.ctx.print_units(), false).into()
-    }
-
     pub fn help(&self) -> JsValue {
-        self.format(&help_markup(), true).into()
+        self.format(&basic_help_markup(), true).into()
+    }
+
+    /// Try to run the input as a command.
+    /// Returns a CommandResult indicating whether the input was a command,
+    /// and if so, what output it produced and any side effects needed.
+    pub fn try_run_command(&mut self, input: &str) -> CommandResult {
+        let mut output = String::new();
+        let mut should_clear = false;
+
+        let result = {
+            let mut cmd_runner = CommandRunner::<()>::new()
+                .print_with(|markup: &m::Markup| {
+                    let fmt = JqueryTerminalFormatter {};
+                    output.push_str(&fmt.format(markup, true));
+                })
+                .enable_clear(|_| {
+                    should_clear = true;
+                    CommandControlFlow::Continue
+                })
+                .enable_reset();
+
+            cmd_runner.try_run_command(input, &mut self.ctx, &mut ())
+        };
+
+        match result {
+            Ok(control_flow) => CommandResult {
+                is_command: control_flow != CommandControlFlow::NotACommand,
+                should_clear,
+                should_reset: control_flow == CommandControlFlow::Reset,
+                output,
+            },
+            Err(err) => CommandResult {
+                is_command: true,
+                should_clear: false,
+                should_reset: false,
+                output: self.format_command_error(&*err),
+            },
+        }
+    }
+
+    fn format_command_error(&self, error: &numbat::command::CommandError) -> String {
+        use codespan_reporting::term::{self, Config};
+
+        let mut writer: Box<dyn BufferedWriter> = match self.format_type {
+            FormatType::JqueryTerminal => Box::new(JqueryTerminalWriter::new()),
+            FormatType::Html => Box::new(HtmlWriter::new()),
+        };
+        let config = Config::default();
+
+        let resolver = self.ctx.resolver();
+        let diagnostic = ResolverDiagnostic { resolver, error };
+
+        for diag in diagnostic.diagnostics() {
+            term::emit(&mut writer, &config, &resolver.files, &diag).unwrap();
+        }
+
+        writer.to_string()
     }
 
     pub fn print_info(&mut self, keyword: &str) -> JsValue {
