@@ -1,3 +1,5 @@
+use std::num::NonZeroUsize;
+
 use crate::markup as m;
 use crate::resolver::ModulePathBorrowed;
 use crate::span::Span;
@@ -5,13 +7,13 @@ use crate::{
     arithmetic::Exponent, decorator::Decorator, markup::Markup, number::Number, prefix::Prefix,
     pretty_print::PrettyPrint,
 };
-use compact_str::{format_compact, CompactString, ToCompactString};
+use compact_str::{CompactString, ToCompactString, format_compact};
 use itertools::Itertools;
 use num_traits::Signed;
 
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub enum UnaryOperator {
-    Factorial,
+    Factorial(NonZeroUsize),
     Negate,
     LogicalNeg,
 }
@@ -147,23 +149,17 @@ impl Expression<'_> {
 
 #[cfg(test)]
 macro_rules! scalar {
-    ( $num:expr ) => {{
-        crate::ast::Expression::Scalar(Span::dummy(), Number::from_f64($num))
-    }};
+    ( $num:expr ) => {{ crate::ast::Expression::Scalar(Span::dummy(), Number::from_f64($num)) }};
 }
 
 #[cfg(test)]
 macro_rules! identifier {
-    ( $name:expr ) => {{
-        crate::ast::Expression::Identifier(Span::dummy(), $name.into())
-    }};
+    ( $name:expr ) => {{ crate::ast::Expression::Identifier(Span::dummy(), $name.into()) }};
 }
 
 #[cfg(test)]
 macro_rules! boolean {
-    ( $name:expr ) => {{
-        crate::ast::Expression::Boolean(Span::dummy(), $name.into())
-    }};
+    ( $name:expr ) => {{ crate::ast::Expression::Boolean(Span::dummy(), $name.into()) }};
 }
 
 #[cfg(test)]
@@ -190,9 +186,9 @@ macro_rules! negate {
 
 #[cfg(test)]
 macro_rules! factorial {
-    ( $lhs:expr ) => {{
+    ( $lhs:expr, $order:expr ) => {{
         crate::ast::Expression::UnaryOperator {
-            op: UnaryOperator::Factorial,
+            op: UnaryOperator::Factorial(NonZeroUsize::new($order).unwrap()),
             expr: Box::new($lhs),
             span_op: Span::dummy(),
         }
@@ -402,6 +398,17 @@ pub enum ProcedureKind {
     Type,
 }
 
+impl ProcedureKind {
+    pub(crate) fn name(&self) -> &'static str {
+        match self {
+            ProcedureKind::Print => "print",
+            ProcedureKind::Assert => "assert",
+            ProcedureKind::AssertEq => "assert_eq",
+            ProcedureKind::Type => "type",
+        }
+    }
+}
+
 #[derive(Debug, Clone, PartialEq)]
 pub enum TypeParameterBound {
     Dim,
