@@ -7,7 +7,7 @@ use crate::ast::ProcedureKind;
 use crate::decorator::Decorator;
 use crate::dimension::DimensionRegistry;
 use crate::interpreter::{
-    Interpreter, InterpreterResult, InterpreterSettings, Result, RuntimeError,
+    Interpreter, InterpreterResult, InterpreterSettings, Result, RuntimeError, RuntimeErrorKind,
 };
 use crate::name_resolution::LAST_RESULT_IDENTIFIERS;
 use crate::prefix::Prefix;
@@ -20,7 +20,7 @@ use crate::unit::{CanonicalName, Unit};
 use crate::unit_registry::{UnitMetadata, UnitRegistry};
 use crate::value::{FunctionReference, Value};
 use crate::vm::{Constant, ExecutionContext, Op, Vm};
-use crate::{decorator, ffi, Type};
+use crate::{Type, decorator};
 
 #[derive(Debug, Clone, Default)]
 pub struct LocalMetadata {
@@ -32,8 +32,7 @@ pub struct LocalMetadata {
 
 #[derive(Debug, Clone)]
 pub struct Local {
-    identifier: CompactString,
-    depth: usize,
+    identifiers: Vec<CompactString>,
     pub metadata: LocalMetadata,
 }
 
@@ -49,29 +48,35 @@ pub struct BytecodeInterpreter {
 }
 
 impl BytecodeInterpreter {
-    fn compile_expression(&mut self, expr: &Expression) -> Result<()> {
+    /// Return a runtime error on the current instruction.
+    pub fn runtime_error(&self, kind: RuntimeErrorKind) -> RuntimeError {
+        self.vm.runtime_error(kind)
+    }
+
+    fn compile_expression(&mut self, expr: &Expression) {
         match expr {
-            Expression::Scalar(_span, n, _type) => {
+            Expression::Scalar(span, n, _type) => {
                 let index = self.vm.add_constant(Constant::Scalar(n.to_f64()));
-                self.vm.add_op1(Op::LoadConstant, index);
+                self.vm.add_op1(Op::LoadConstant, index, *span);
             }
-            Expression::Identifier(_span, identifier, _type) => {
+            Expression::Identifier(span, identifier, _type) => {
                 // Searching in reverse order ensures that we find the innermost identifier of that name first (shadowing)
 
                 let current_depth = self.locals.len() - 1;
 
                 if let Some(position) = self.locals[current_depth]
                     .iter()
-                    .rposition(|l| l.identifier == identifier && l.depth == current_depth)
+                    .rposition(|l| l.identifiers.iter().any(|n| n == identifier))
                 {
-                    self.vm.add_op1(Op::GetLocal, position as u16); // TODO: check overflow
+                    self.vm.add_op1(Op::GetLocal, position as u16, *span); // TODO: check overflow
                 } else if let Some(upvalue_position) = self.locals[0]
                     .iter()
-                    .rposition(|l| l.identifier == identifier)
+                    .rposition(|l| l.identifiers.iter().any(|n| n == identifier))
                 {
-                    self.vm.add_op1(Op::GetUpvalue, upvalue_position as u16);
+                    self.vm
+                        .add_op1(Op::GetUpvalue, upvalue_position as u16, *span);
                 } else if LAST_RESULT_IDENTIFIERS.contains(identifier) {
-                    self.vm.add_op(Op::GetLastResult);
+                    self.vm.add_op(Op::GetLastResult, *span);
                 } else if let Some(is_foreign) = self.functions.get(*identifier) {
                     let index = self
                         .vm
@@ -80,39 +85,39 @@ impl BytecodeInterpreter {
                         } else {
                             FunctionReference::Normal(identifier.to_compact_string())
                         }));
-                    self.vm.add_op1(Op::LoadConstant, index);
+                    self.vm.add_op1(Op::LoadConstant, index, *span);
                 } else {
                     unreachable!("Unknown identifier '{identifier}'")
                 }
             }
-            Expression::UnitIdentifier(_span, prefix, unit_name, _full_name, _type) => {
+            Expression::UnitIdentifier(span, prefix, unit_name, _full_name, _type) => {
                 let index = self
                     .unit_name_to_constant_index
                     .get(unit_name)
                     .expect("unit should already exist");
 
-                self.vm.add_op1(Op::LoadConstant, *index);
+                self.vm.add_op1(Op::LoadConstant, *index, *span);
 
                 if prefix != &Prefix::none() {
                     let prefix_idx = self.vm.add_prefix(*prefix);
-                    self.vm.add_op1(Op::ApplyPrefix, prefix_idx);
+                    self.vm.add_op1(Op::ApplyPrefix, prefix_idx, *span);
                 }
             }
-            Expression::UnaryOperator(_span, UnaryOperator::Negate, rhs, _type) => {
-                self.compile_expression(rhs)?;
-                self.vm.add_op(Op::Negate);
+            Expression::UnaryOperator(span, UnaryOperator::Negate, rhs, _type) => {
+                self.compile_expression(rhs);
+                self.vm.add_op(Op::Negate, *span);
             }
-            Expression::UnaryOperator(_span, UnaryOperator::Factorial, lhs, _type) => {
-                self.compile_expression(lhs)?;
-                self.vm.add_op(Op::Factorial);
+            Expression::UnaryOperator(span, UnaryOperator::Factorial(order), lhs, _type) => {
+                self.compile_expression(lhs);
+                self.vm.add_op1(Op::Factorial, order.get() as u16, *span);
             }
-            Expression::UnaryOperator(_span, UnaryOperator::LogicalNeg, lhs, _type) => {
-                self.compile_expression(lhs)?;
-                self.vm.add_op(Op::LogicalNeg);
+            Expression::UnaryOperator(span, UnaryOperator::LogicalNeg, lhs, _type) => {
+                self.compile_expression(lhs);
+                self.vm.add_op(Op::LogicalNeg, *span);
             }
-            Expression::BinaryOperator(_span, operator, lhs, rhs, _type) => {
-                self.compile_expression(lhs)?;
-                self.compile_expression(rhs)?;
+            Expression::BinaryOperator(span, operator, lhs, rhs, _type) => {
+                self.compile_expression(lhs);
+                self.compile_expression(rhs);
 
                 let op = match operator {
                     BinaryOperator::Add => Op::Add,
@@ -130,11 +135,16 @@ impl BytecodeInterpreter {
                     BinaryOperator::LogicalAnd => Op::LogicalAnd,
                     BinaryOperator::LogicalOr => Op::LogicalOr,
                 };
-                self.vm.add_op(op);
+                self.vm.add_op(
+                    op,
+                    span.unwrap_or_else(|| {
+                        crate::span::Span::in_between(lhs.full_span(), rhs.full_span())
+                    }),
+                );
             }
-            Expression::BinaryOperatorForDate(_span, operator, lhs, rhs, type_) => {
-                self.compile_expression(lhs)?;
-                self.compile_expression(rhs)?;
+            Expression::BinaryOperatorForDate(span, operator, lhs, rhs, type_) => {
+                self.compile_expression(lhs);
+                self.compile_expression(rhs);
 
                 // if the result is a duration:
                 let op = if type_.is_dtype() {
@@ -145,7 +155,13 @@ impl BytecodeInterpreter {
                     // this specific purpose. We also need to handle errors in case no such unit
                     // was registered.
                     let second_idx = self.unit_name_to_constant_index.get("second");
-                    self.vm.add_op1(Op::LoadConstant, *second_idx.unwrap());
+                    self.vm.add_op1(
+                        Op::LoadConstant,
+                        *second_idx.unwrap(),
+                        span.unwrap_or_else(|| {
+                            crate::span::Span::in_between(lhs.full_span(), rhs.full_span())
+                        }),
+                    );
                     Op::DiffDateTime
                 } else {
                     match operator {
@@ -155,24 +171,31 @@ impl BytecodeInterpreter {
                     }
                 };
 
-                self.vm.add_op(op);
+                self.vm.add_op(
+                    op,
+                    span.unwrap_or_else(|| {
+                        crate::span::Span::in_between(lhs.full_span(), rhs.full_span())
+                    }),
+                );
             }
-            Expression::FunctionCall(_span, _full_span, name, args, _type) => {
+            Expression::FunctionCall(_span, full_span, name, args, _type) => {
                 // Put all arguments on top of the stack
                 for arg in args {
-                    self.compile_expression(arg)?;
+                    self.compile_expression(arg);
                 }
 
                 if let Some(idx) = self.vm.get_ffi_callable_idx(name) {
                     // TODO: check overflow:
-                    self.vm.add_op2(Op::FFICallFunction, idx, args.len() as u16);
+                    self.vm
+                        .add_op2(Op::FFICallFunction, idx, args.len() as u16, *full_span);
                 } else {
                     let idx = self.vm.get_function_idx(name);
 
-                    self.vm.add_op2(Op::Call, idx, args.len() as u16); // TODO: check overflow
+                    self.vm
+                        .add_op2(Op::Call, idx, args.len() as u16, *full_span); // TODO: check overflow
                 }
             }
-            Expression::InstantiateStruct(_span, exprs, struct_info) => {
+            Expression::InstantiateStruct(span, exprs, struct_info) => {
                 // structs must be consistently ordered in the VM, so we reorder
                 // the field values so that they are evaluated in the order the
                 // struct fields are defined.
@@ -182,16 +205,20 @@ impl BytecodeInterpreter {
                     .sorted_by_key(|(n, _)| struct_info.fields.get_index_of(*n).unwrap());
 
                 for (_, expr) in sorted_exprs.rev() {
-                    self.compile_expression(expr)?;
+                    self.compile_expression(expr);
                 }
 
                 let struct_info_idx = self.vm.get_structinfo_idx(&struct_info.name).unwrap() as u16;
 
-                self.vm
-                    .add_op2(Op::BuildStructInstance, struct_info_idx, exprs.len() as u16);
+                self.vm.add_op2(
+                    Op::BuildStructInstance,
+                    struct_info_idx,
+                    exprs.len() as u16,
+                    *span,
+                );
             }
-            Expression::AccessField(_span, _full_span, expr, attr, struct_type, _result_type) => {
-                self.compile_expression(expr)?;
+            Expression::AccessField(_span, full_span, expr, attr, struct_type, _result_type) => {
+                self.compile_expression(expr);
 
                 let Type::Struct(ref struct_info) = struct_type.to_concrete_type() else {
                     unreachable!(
@@ -201,108 +228,103 @@ impl BytecodeInterpreter {
 
                 let idx = struct_info.fields.get_index_of(*attr).unwrap();
 
-                self.vm.add_op1(Op::AccessStructField, idx as u16);
+                self.vm
+                    .add_op1(Op::AccessStructField, idx as u16, *full_span);
             }
-            Expression::CallableCall(_span, callable, args, _type) => {
+            Expression::CallableCall(span, callable, args, _type) => {
                 // Put all arguments on top of the stack
                 for arg in args {
-                    self.compile_expression(arg)?;
+                    self.compile_expression(arg);
                 }
 
                 // Put the callable on top of the stack
-                self.compile_expression(callable)?;
+                self.compile_expression(callable);
 
-                self.vm.add_op1(Op::CallCallable, args.len() as u16);
+                self.vm.add_op1(Op::CallCallable, args.len() as u16, *span);
             }
-            Expression::Boolean(_, val) => {
+            Expression::Boolean(span, val) => {
                 let index = self.vm.add_constant(Constant::Boolean(*val));
-                self.vm.add_op1(Op::LoadConstant, index);
+                self.vm.add_op1(Op::LoadConstant, index, *span);
             }
-            Expression::String(_, string_parts) => {
+            Expression::String(span, string_parts) => {
                 for part in string_parts {
                     match part {
                         StringPart::Fixed(s) => {
                             let index = self.vm.add_constant(Constant::String(s.clone()));
-                            self.vm.add_op1(Op::LoadConstant, index)
+                            self.vm.add_op1(Op::LoadConstant, index, *span)
                         }
                         StringPart::Interpolation {
                             expr,
                             span: _,
                             format_specifiers,
                         } => {
-                            self.compile_expression(expr)?;
+                            self.compile_expression(expr);
                             let index = self.vm.add_constant(Constant::FormatSpecifiers(
                                 format_specifiers.map(|s| s.to_compact_string()),
                             ));
-                            self.vm.add_op1(Op::LoadConstant, index)
+                            self.vm.add_op1(Op::LoadConstant, index, *span)
                         }
                     }
                 }
-                self.vm.add_op1(Op::JoinString, string_parts.len() as u16); // TODO: this can overflow
+                self.vm
+                    .add_op1(Op::JoinString, string_parts.len() as u16, *span); // TODO: this can overflow
             }
-            Expression::Condition(_, condition, then_expr, else_expr) => {
-                self.compile_expression(condition)?;
+            Expression::Condition(span, condition, then_expr, else_expr) => {
+                self.compile_expression(condition);
 
                 let if_jump_offset = self.vm.current_offset() + 1; // +1 for the opcode
-                self.vm.add_op1(Op::JumpIfFalse, 0xffff);
+                self.vm.add_op1(Op::JumpIfFalse, 0xffff, *span);
 
-                self.compile_expression(then_expr)?;
+                self.compile_expression(then_expr);
 
                 let else_jump_offset = self.vm.current_offset() + 1;
-                self.vm.add_op1(Op::Jump, 0xffff);
+                self.vm.add_op1(Op::Jump, 0xffff, *span);
 
                 let else_block_offset = self.vm.current_offset();
                 self.vm
                     .patch_u16_value_at(if_jump_offset, else_block_offset - (if_jump_offset + 2));
 
-                self.compile_expression(else_expr)?;
+                self.compile_expression(else_expr);
 
                 let end_offset = self.vm.current_offset();
 
                 self.vm
                     .patch_u16_value_at(else_jump_offset, end_offset - (else_jump_offset + 2));
             }
-            Expression::List(_, elements, _) => {
+            Expression::List(span, elements, _) => {
                 for element in elements {
-                    self.compile_expression(element)?;
+                    self.compile_expression(element);
                 }
 
-                self.vm.add_op1(Op::BuildList, elements.len() as u16);
+                self.vm.add_op1(Op::BuildList, elements.len() as u16, *span);
             }
             Expression::TypedHole(_, _) => {
                 unreachable!("Typed holes cause type inference errors")
             }
         };
-
-        Ok(())
     }
 
-    fn compile_define_variable(&mut self, define_variable: &DefineVariable) -> Result<()> {
+    fn compile_define_variable(&mut self, define_variable: &DefineVariable) {
         let DefineVariable(identifier, decorators, expr, _annotation, _type, _readable_type) =
             define_variable;
         let current_depth = self.current_depth();
 
         // For variables, we ignore the prefix info and only use the names
-        let aliases = crate::decorator::name_and_aliases(identifier, decorators)
+        let identifiers = crate::decorator::name_and_aliases(identifier, decorators)
             .map(|(name, _)| name.to_compact_string())
             .collect::<Vec<_>>();
         let metadata = LocalMetadata {
             name: crate::decorator::name(decorators).map(CompactString::from),
             url: crate::decorator::url(decorators).map(CompactString::from),
             description: crate::decorator::description(decorators),
-            aliases: aliases.clone(),
+            aliases: identifiers.clone(),
         };
 
-        for alias_name in aliases {
-            self.compile_expression(expr)?;
-
-            self.locals[current_depth].push(Local {
-                identifier: alias_name.clone(),
-                depth: current_depth,
-                metadata: metadata.clone(),
-            });
-        }
-        Ok(())
+        self.compile_expression(expr);
+        self.locals[current_depth].push(Local {
+            identifiers,
+            metadata,
+        });
     }
 
     fn compile_statement(
@@ -312,11 +334,11 @@ impl BytecodeInterpreter {
     ) -> Result<()> {
         match stmt {
             Statement::Expression(expr) => {
-                self.compile_expression(expr)?;
-                self.vm.add_op(Op::Return);
+                self.compile_expression(expr);
+                self.vm.add_op(Op::Return, expr.full_span());
             }
             Statement::DefineVariable(define_variable) => {
-                self.compile_define_variable(define_variable)?
+                self.compile_define_variable(define_variable);
             }
             Statement::DefineFunction(
                 name,
@@ -336,18 +358,17 @@ impl BytecodeInterpreter {
                 let current_depth = self.current_depth();
                 for parameter in parameters {
                     self.locals[current_depth].push(Local {
-                        identifier: parameter.1.to_compact_string(),
-                        depth: current_depth,
+                        identifiers: [parameter.1.to_compact_string()].into(),
                         metadata: LocalMetadata::default(),
                     });
                 }
                 for local_variables in local_variables {
-                    self.compile_define_variable(local_variables)?;
+                    self.compile_define_variable(local_variables);
                 }
 
-                self.compile_expression(expr)?;
+                self.compile_expression(expr);
 
-                self.vm.add_op(Op::Return);
+                self.vm.add_op(Op::Return, expr.full_span());
 
                 self.locals.pop();
 
@@ -404,7 +425,10 @@ impl BytecodeInterpreter {
                             metric_prefixes: decorators.contains(&Decorator::MetricPrefixes),
                         },
                     )
-                    .map_err(RuntimeError::UnitRegistryError)?;
+                    .map_err(|e| {
+                        self.vm
+                            .runtime_error(RuntimeErrorKind::UnitRegistryError(e))
+                    })?;
 
                 let constant_idx = self.vm.add_constant(Constant::Unit(Unit::new_base(
                     unit_name.to_compact_string(),
@@ -417,6 +441,7 @@ impl BytecodeInterpreter {
             }
             Statement::DefineDerivedUnit(
                 unit_name,
+                full_span,
                 expr,
                 decorators,
                 annotation,
@@ -436,9 +461,6 @@ impl BytecodeInterpreter {
                 ))); // TODO: dummy is just a temp. value until the SetUnitConstant op runs
                 let unit_information_idx = self.vm.add_unit_information(
                     unit_name,
-                    Some(
-                        &crate::decorator::get_canonical_unit_name(unit_name, &decorators[..]).name,
-                    ),
                     UnitMetadata {
                         type_: type_.to_concrete_type(), // We guarantee that derived-unit definitions do not contain generics, so no TGen(..)s can escape
                         readable_type: annotation
@@ -455,9 +477,13 @@ impl BytecodeInterpreter {
                     },
                 ); // TODO: there is some asymmetry here because we do not introduce identifiers for base units
 
-                self.compile_expression(expr)?;
-                self.vm
-                    .add_op2(Op::SetUnitConstant, unit_information_idx, constant_idx);
+                self.compile_expression(expr);
+                self.vm.add_op2(
+                    Op::SetUnitConstant,
+                    unit_information_idx,
+                    constant_idx,
+                    *full_span,
+                );
 
                 // TODO: code duplication with DeclareBaseUnit branch above
                 for (name, _) in decorator::name_and_aliases(unit_name, decorators) {
@@ -465,7 +491,7 @@ impl BytecodeInterpreter {
                         .insert(name.into(), constant_idx);
                 }
             }
-            Statement::ProcedureCall(ProcedureKind::Type, args) => {
+            Statement::ProcedureCall(ProcedureKind::Type, full_span, args) => {
                 assert_eq!(args.len(), 1);
                 let arg = &args[0];
 
@@ -473,15 +499,15 @@ impl BytecodeInterpreter {
                 let idx = self.vm.add_string(
                     m::dimmed("=") + m::whitespace(" ") + arg.get_type_scheme().pretty_print(), // TODO
                 );
-                self.vm.add_op1(Op::PrintString, idx);
+                self.vm.add_op1(Op::PrintString, idx, *full_span);
             }
-            Statement::ProcedureCall(kind, args) => {
+            Statement::ProcedureCall(kind, full_span, args) => {
                 // Put all arguments on top of the stack
                 for arg in args {
-                    self.compile_expression(arg)?;
+                    self.compile_expression(arg);
                 }
 
-                let name = &ffi::procedures().get(kind).unwrap().name;
+                let name = kind.name();
 
                 let callable_idx = self.vm.get_ffi_callable_idx(name).unwrap();
 
@@ -493,6 +519,7 @@ impl BytecodeInterpreter {
                     callable_idx,
                     args.len() as u16,
                     spans_idx,
+                    *full_span,
                 );
                 // TODO: check overflow
             }
@@ -544,7 +571,9 @@ impl BytecodeInterpreter {
     }
 
     pub fn lookup_global(&self, name: &str) -> Option<&Local> {
-        self.locals[0].iter().find(|l| l.identifier == name)
+        self.locals[0]
+            .iter()
+            .find(|l| l.identifiers.iter().any(|n| n == name))
     }
 }
 
