@@ -3,14 +3,14 @@ use std::str::{FromStr, SplitWhitespace};
 use compact_str::ToCompactString;
 
 use crate::{
-    diagnostic::ErrorDiagnostic,
-    help::help_markup,
+    Context, ParseError, RuntimeError,
+    diagnostic::{ErrorDiagnostic, ResolverDiagnostic},
+    help::basic_help_markup,
     markup::{self as m, Markup},
     parser::ParseErrorKind,
     resolver::CodeSource,
     session_history::{SessionHistory, SessionHistoryOptions},
     span::{ByteIndex, Span},
-    Context, ParseError, RuntimeError,
 };
 
 #[derive(Debug, Clone, PartialEq)]
@@ -21,12 +21,15 @@ pub enum ListItems {
     Units,
 }
 
-enum QuitAlias {
+#[derive(Debug, Clone, Copy, PartialEq)]
+pub enum QuitAlias {
     Quit,
     Exit,
 }
 
-enum CommandKind {
+#[non_exhaustive]
+#[derive(Debug, Clone, Copy, PartialEq)]
+pub enum CommandKind {
     Help,
     Info,
     List,
@@ -55,11 +58,18 @@ impl FromStr for CommandKind {
     }
 }
 
+#[derive(Debug, Clone, Copy, PartialEq)]
+enum HelpKind {
+    BasicHelp,
+    AllCommands,
+}
+
 #[derive(Debug, PartialEq, Eq)]
 #[must_use]
 pub enum CommandControlFlow {
     Continue,
     Return,
+    Reset,
     NotACommand,
 }
 
@@ -81,105 +91,75 @@ impl From<RuntimeError> for CommandError {
     }
 }
 
-impl ErrorDiagnostic for CommandError {
+impl ErrorDiagnostic for ResolverDiagnostic<'_, CommandError> {
     fn diagnostics(&self) -> Vec<crate::Diagnostic> {
-        match self {
+        match self.error {
             CommandError::Parse(parse_error) => parse_error.diagnostics(),
-            CommandError::Runtime(runtime_error) => runtime_error.diagnostics(),
+            CommandError::Runtime(runtime_error) => ResolverDiagnostic {
+                resolver: self.resolver,
+                error: runtime_error,
+            }
+            .diagnostics(),
         }
     }
 }
 
-enum Command<'session, 'input, Editor> {
+enum ParsedCommand<'session, 'input> {
     Help {
-        print_fn: fn(&Markup),
+        help_kind: HelpKind,
     },
     Info {
         item: &'input str,
-        print_fn: fn(&Markup),
     },
     List {
         items: Option<ListItems>,
-        print_fn: fn(&Markup),
     },
-    Clear {
-        clear_fn: fn(&mut Editor) -> CommandControlFlow,
-    },
-    Save(SaveCmdArgs<'session, 'input>),
-    Reset {
-        ctx_ctor: fn() -> Context,
-        clear_fn: Option<fn(&mut Editor) -> CommandControlFlow>,
+    Clear,
+    Save {
+        session_history: &'session SessionHistory,
+        dst: &'input str,
     },
+    Reset,
     Quit,
 }
 
-struct SaveCmdArgs<'session, 'input> {
-    session_history: &'session SessionHistory,
-    dst: &'input str,
-    print_fn: Option<fn(&Markup)>,
-}
-
-impl SaveCmdArgs<'_, '_> {
-    fn save(&self) -> Result<(), Box<RuntimeError>> {
-        let Self {
-            session_history,
-            dst,
-            print_fn,
-        } = self;
-
-        session_history.save(
-            dst,
-            SessionHistoryOptions {
-                include_err_lines: false,
-                trim_lines: true,
-            },
-        )?;
-
-        if let Some(print_markup) = print_fn {
-            let markup = m::text("successfully saved session history to")
-                + m::space()
-                + m::string(dst.to_compact_string());
-            print_markup(&markup)
-        }
-
-        Ok(())
-    }
-}
-
-pub struct CommandRunner<Editor = ()> {
-    print_markup: Option<fn(&Markup)>,
-    clear: Option<fn(&mut Editor) -> CommandControlFlow>,
+pub struct CommandRunner<'a, Editor = ()> {
+    print_markup: Option<Box<dyn FnMut(&Markup) + 'a>>,
+    clear: Option<Box<dyn FnMut(&mut Editor) -> CommandControlFlow + 'a>>,
     session_history: Option<SessionHistory>,
-    ctx_ctor: Option<fn() -> Context>,
+    reset: Option<()>,
     quit: Option<()>,
 }
 
 // cannot be derived because `#[derive(Default)]` introduces the bound `Editor:
 // Default`, which is not necessary
-impl<Editor> Default for CommandRunner<Editor> {
+impl<Editor> Default for CommandRunner<'_, Editor> {
     fn default() -> Self {
         Self {
             print_markup: None,
             clear: None,
             session_history: None,
-            ctx_ctor: None,
+            reset: None,
             quit: None,
         }
     }
 }
 
-impl<Editor> CommandRunner<Editor> {
+impl<'a, Editor> CommandRunner<'a, Editor> {
     pub fn new() -> Self {
         Self::default()
     }
 
-    pub fn print_with(mut self, action: fn(&Markup)) -> Self {
-        self.print_markup = Some(action);
+    pub fn print_with(mut self, action: impl FnMut(&Markup) + 'a) -> Self {
+        self.print_markup = Some(Box::new(action));
         self
     }
 
-    pub fn enable_clear(mut self, action: fn(&mut Editor) -> CommandControlFlow) -> Self {
-        self.clear = Some(action);
+    pub fn enable_clear(
+        mut self,
+        action: impl FnMut(&mut Editor) -> CommandControlFlow + 'a,
+    ) -> Self {
+        self.clear = Some(Box::new(action));
         self
     }
 
@@ -188,8 +168,8 @@ impl<Editor> CommandRunner<Editor> {
         self
     }
 
-    pub fn enable_reset(mut self, ctx_ctor: fn() -> Context) -> Self {
-        self.ctx_ctor = Some(ctx_ctor);
+    pub fn enable_reset(mut self) -> Self {
+        self.reset = Some(());
         self
     }
 
@@ -205,11 +185,150 @@ impl<Editor> CommandRunner<Editor> {
         session_history.push(line.to_compact_string(), result);
     }
 
-    fn get_command<'a, 'b>(
-        &'a self,
+    fn help_markup(&self, help_kind: HelpKind) -> m::Markup {
+        match help_kind {
+            HelpKind::BasicHelp => basic_help_markup(),
+            HelpKind::AllCommands => self.all_commands_markup(),
+        }
+    }
+
+    fn all_commands_markup(&self) -> m::Markup {
+        fn m_cmd(cmd: &'static str) -> m::Markup {
+            m::keyword(cmd)
+        }
+
+        fn m_arg(arg: &'static str) -> m::Markup {
+            m::value(arg)
+        }
+
+        // Column width for command + args (not including indent)
+        const CMD_COLUMN_WIDTH: usize = 18;
+
+        fn cmd(
+            cmd: &'static str,
+            args: impl AsRef<[&'static str]>,
+            aliases: impl AsRef<[&'static str]>,
+            help: &'static str,
+        ) -> m::Markup {
+            cmd_fmt(cmd, args, aliases, m::text(help))
+        }
+
+        fn cmd_fmt(
+            cmd: &'static str,
+            args: impl AsRef<[&'static str]>,
+            aliases: impl AsRef<[&'static str]>,
+            help: m::Markup,
+        ) -> m::Markup {
+            let indent = m::text("  ");
+            let mut output = indent;
+            output += m_cmd(cmd);
+
+            // Calculate the width of command + args
+            let args = args.as_ref();
+            let mut cmd_width = cmd.len();
+            for arg in args {
+                output += m::space();
+                output += m_arg(arg);
+                cmd_width += 1 + arg.len(); // space + arg
+            }
+
+            // Pad to align descriptions
+            let padding = CMD_COLUMN_WIDTH.saturating_sub(cmd_width);
+            let padding_str = " ".repeat(padding).to_compact_string();
+            output += m::whitespace(padding_str);
+            output += help;
+
+            let aliases = aliases.as_ref();
+            let mut aliases_iter = aliases.iter();
+            if let Some(first_alias) = aliases_iter.next() {
+                if aliases.len() == 1 {
+                    output += m::text(" (alias: ");
+                } else {
+                    output += m::text(" (aliases: ");
+                }
+                output += m_cmd(first_alias);
+
+                for alias in aliases_iter {
+                    output += m::text(", ");
+                    output += m_cmd(alias);
+                }
+                output += m::text(")");
+            }
+
+            output += m::nl();
+
+            output
+        }
+
+        let mut output = m::nl()
+            + cmd("help", [], ["?"], "show a basic introduction to Numbat")
+            + cmd(
+                "help",
+                ["commands"],
+                [],
+                "brief description of all available commands",
+            )
+            + cmd(
+                "info",
+                ["<identifier>"],
+                [],
+                "get more information about functions, variables, units or dimensions",
+            )
+            + cmd("list", [], [], "show all currently defined items")
+            + cmd_fmt(
+                "list",
+                ["<what>"],
+                [],
+                m::text("show all currently defined ")
+                    + m_arg("functions")
+                    + m::text(", ")
+                    + m_arg("definitions")
+                    + m::text(", ")
+                    + m_arg("variables")
+                    + m::text(", or ")
+                    + m_arg("units"),
+            );
+
+        if self.session_history.is_some() {
+            output += cmd_fmt(
+                "save",
+                [],
+                [],
+                m::text("save the current session history to ")
+                    + m::string("history.nbt")
+                    + m::text(" in the current directory"),
+            );
+            output += cmd_fmt(
+                "save",
+                ["<path>"],
+                [],
+                m::text("save the current session history to file")
+                    + m::text("(the recommended file extension is ")
+                    + m::string(".nbt")
+                    + m::text(")"),
+            );
+        }
+
+        if self.clear.is_some() {
+            output += cmd("clear", [], [], "clear the console output");
+        }
+
+        if self.reset.is_some() {
+            output += cmd("reset", [], [], "completely reset the interpreter state");
+        }
+
+        if self.quit.is_some() {
+            output += cmd("exit", [], ["quit"], "close this session");
+        }
+
+        output
+    }
+
+    fn parse_command<'s, 'b>(
+        &'s self,
         line: &'b str,
         ctx: &mut Context,
-    ) -> Result<Option<Command<'a, 'b, Editor>>, Box<CommandError>> {
+    ) -> Result<Option<ParsedCommand<'s, 'b>>, Box<CommandError>> {
         let Some(mut parser) = CommandParser::new(
             line,
             ctx.resolver_mut().add_code_source(CodeSource::Text, line),
@@ -221,7 +340,7 @@ impl<Editor> CommandRunner<Editor> {
             print_markup,
             clear,
             session_history,
-            ctx_ctor,
+            reset,
             quit,
         } = self;
 
@@ -229,20 +348,41 @@ impl<Editor> CommandRunner<Editor> {
         // with an if-let guard. https://github.com/rust-lang/rust/issues/51114
         Ok(Some(match &parser.command_kind {
             CommandKind::Help => {
-                let &Some(print_fn) = print_markup else {
+                if print_markup.is_none() {
                     return Ok(None);
+                }
+
+                let help_arg = parser.args.next();
+
+                if parser.args.next().is_some() {
+                    return Err(Box::new(
+                        parser
+                            .err_through_end_from(2, "`help` takes at most one argument")
+                            .into(),
+                    ));
+                }
+
+                let help_kind = match help_arg {
+                    None => HelpKind::BasicHelp,
+                    Some("commands") => HelpKind::AllCommands,
+                    _ => {
+                        return Err(Box::new(
+                            parser
+                                .err_at_idx(
+                                    1,
+                                    "if provided, the argument to `help` must be \"commands\"",
+                                )
+                                .into(),
+                        ));
+                    }
                 };
 
-                parser
-                    .ensure_zero_args("help", "; use `info <item>` for information about an item")
-                    .map_err(|err| Box::new(err.into()))?;
-
-                Command::Help { print_fn }
+                ParsedCommand::Help { help_kind }
             }
             CommandKind::Info => {
-                let &Some(print_fn) = print_markup else {
+                if print_markup.is_none() {
                     return Ok(None);
-                };
+                }
                 let err_msg = "`info` requires exactly one argument, the item to get info on";
                 let Some(item) = parser.args.next() else {
                     return Err(Box::new(parser.err_at_idx(0, err_msg).into()));
@@ -252,12 +392,12 @@ impl<Editor> CommandRunner<Editor> {
                     return Err(Box::new(parser.err_through_end_from(1, err_msg).into()));
                 }
 
-                Command::Info { item, print_fn }
+                ParsedCommand::Info { item }
             }
             CommandKind::List => {
-                let &Some(print_fn) = print_markup else {
+                if print_markup.is_none() {
                     return Ok(None);
-                };
+                }
 
                 let items = parser.args.next();
 
@@ -288,26 +428,24 @@ impl<Editor> CommandRunner<Editor> {
                     }
                 };
 
-                Command::List { items, print_fn }
+                ParsedCommand::List { items }
             }
             CommandKind::Clear => {
-                let &Some(clear_fn) = clear else {
+                if clear.is_none() {
                     return Ok(None);
-                };
+                }
 
                 parser
                     .ensure_zero_args("clear", "")
                     .map_err(|err| Box::new(err.into()))?;
 
-                Command::Clear { clear_fn }
+                ParsedCommand::Clear
             }
             CommandKind::Save => {
                 let Some(session_history) = session_history else {
                     return Ok(None);
                 };
 
-                let print_fn = print_markup.as_ref().copied();
-
                 let dst = parser.args.next().unwrap_or("history.nbt");
 
                 if parser.args.next().is_some() {
@@ -321,30 +459,26 @@ impl<Editor> CommandRunner<Editor> {
                     ));
                 }
 
-                Command::Save(SaveCmdArgs {
+                ParsedCommand::Save {
                     session_history,
                     dst,
-                    print_fn,
-                })
+                }
             }
             CommandKind::Reset => {
-                let &Some(ctx_ctor) = ctx_ctor else {
+                if reset.is_none() {
                     return Ok(None);
-                };
+                }
 
                 parser
                     .ensure_zero_args("reset", "")
                     .map_err(|err| Box::new(err.into()))?;
 
-                Command::Reset {
-                    ctx_ctor,
-                    clear_fn: *clear,
-                }
+                ParsedCommand::Reset
             }
             CommandKind::Quit(quit_alias) => {
-                let Some(_) = quit else {
+                if quit.is_none() {
                     return Ok(None);
-                };
+                }
 
                 parser
                     .ensure_zero_args(
@@ -356,7 +490,7 @@ impl<Editor> CommandRunner<Editor> {
                     )
                     .map_err(|err| Box::new(err.into()))?;
 
-                Command::Quit
+                ParsedCommand::Quit
             }
         }))
     }
@@ -372,25 +506,31 @@ impl<Editor> CommandRunner<Editor> {
     /// that's because it's not a command at all, or it's a disabled command), then this
     /// returns `CommandControlFlow::NotACommand`.
     pub fn try_run_command(
-        &self,
+        &mut self,
         line: &str,
         ctx: &mut Context,
         editor: &mut Editor,
     ) -> Result<CommandControlFlow, Box<CommandError>> {
-        let Some(output) = self.get_command(line, ctx)? else {
+        let Some(parsed) = self.parse_command(line, ctx)? else {
             return Ok(CommandControlFlow::NotACommand);
         };
 
-        Ok(match output {
-            Command::Help { print_fn } => {
-                print_fn(&help_markup());
+        Ok(match parsed {
+            ParsedCommand::Help { help_kind } => {
+                let markup = self.help_markup(help_kind);
+                if let Some(print_fn) = self.print_markup.as_mut() {
+                    print_fn(&markup);
+                }
                 CommandControlFlow::Continue
             }
-            Command::Info { item, print_fn } => {
-                print_fn(&ctx.print_info_for_keyword(item));
+            ParsedCommand::Info { item } => {
+                let markup = ctx.print_info_for_keyword(item);
+                if let Some(print_fn) = self.print_markup.as_mut() {
+                    print_fn(&markup);
+                }
                 CommandControlFlow::Continue
             }
-            Command::List { items, print_fn } => {
+            ParsedCommand::List { items } => {
                 let markup = match items {
                     None => ctx.print_environment(),
                     Some(ListItems::Functions) => ctx.print_functions(),
@@ -398,22 +538,47 @@ impl<Editor> CommandRunner<Editor> {
                     Some(ListItems::Variables) => ctx.print_variables(),
                     Some(ListItems::Units) => ctx.print_units(),
                 };
-                print_fn(&markup);
+                if let Some(print_fn) = self.print_markup.as_mut() {
+                    print_fn(&markup);
+                }
                 CommandControlFlow::Continue
             }
-            Command::Clear { clear_fn } => clear_fn(editor),
-            Command::Save(save_args) => {
-                save_args.save().map_err(|err| Box::new((*err).into()))?;
-                CommandControlFlow::Continue
-            }
-            Command::Reset { ctx_ctor, clear_fn } => {
-                *ctx = ctx_ctor();
-                match clear_fn {
-                    Some(clear_fn) => clear_fn(editor),
-                    None => CommandControlFlow::Continue,
+            ParsedCommand::Clear => {
+                if let Some(clear_fn) = self.clear.as_mut() {
+                    clear_fn(editor)
+                } else {
+                    CommandControlFlow::Continue
                 }
             }
-            Command::Quit => CommandControlFlow::Return,
+            ParsedCommand::Save {
+                session_history,
+                dst,
+            } => {
+                session_history
+                    .save(
+                        dst,
+                        SessionHistoryOptions {
+                            include_err_lines: false,
+                            trim_lines: true,
+                        },
+                    )
+                    .map_err(|err| CommandError::Runtime(ctx.interpreter.runtime_error(*err)))?;
+
+                if let Some(print_fn) = self.print_markup.as_mut() {
+                    let markup = m::text("successfully saved session history to")
+                        + m::space()
+                        + m::string(dst.to_compact_string());
+                    print_fn(&markup);
+                }
+                CommandControlFlow::Continue
+            }
+            ParsedCommand::Reset => {
+                if let Some(clear_fn) = self.clear.as_mut() {
+                    let _ = clear_fn(editor);
+                }
+                CommandControlFlow::Reset
+            }
+            ParsedCommand::Quit => CommandControlFlow::Return,
         })
     }
 }
@@ -532,7 +697,7 @@ mod test {
 
     #[derive(Debug, Clone, PartialEq)]
     pub enum BareCommand<'a> {
-        Help,
+        Help { help_kind: HelpKind },
         Info { item: &'a str },
         List { items: Option<ListItems> },
         Clear,
@@ -541,26 +706,26 @@ mod test {
         Quit,
     }
 
-    impl<'b, Editor> Command<'_, 'b, Editor> {
-        fn into_bare(self) -> BareCommand<'b> {
-            match self {
-                Command::Help { print_fn: _ } => BareCommand::Help,
-                Command::Info { item, .. } => BareCommand::Info { item },
-                Command::List { items, .. } => BareCommand::List { items },
-                Command::Clear { clear_fn: _ } => BareCommand::Clear,
-                Command::Save(SaveCmdArgs { dst, .. }) => BareCommand::Save { dst },
-                Command::Reset { .. } => BareCommand::Reset,
-                Command::Quit => BareCommand::Quit,
+    impl<'a> From<ParsedCommand<'_, 'a>> for BareCommand<'a> {
+        fn from(cmd: ParsedCommand<'_, 'a>) -> Self {
+            match cmd {
+                ParsedCommand::Help { help_kind } => BareCommand::Help { help_kind },
+                ParsedCommand::Info { item } => BareCommand::Info { item },
+                ParsedCommand::List { items } => BareCommand::List { items },
+                ParsedCommand::Clear => BareCommand::Clear,
+                ParsedCommand::Save { dst, .. } => BareCommand::Save { dst },
+                ParsedCommand::Reset => BareCommand::Reset,
+                ParsedCommand::Quit => BareCommand::Quit,
             }
         }
     }
 
-    fn new_runner() -> CommandRunner<()> {
+    fn new_runner() -> CommandRunner<'static, ()> {
         CommandRunner::new()
             .print_with(|_| {})
             .enable_clear(|_| CommandControlFlow::Continue)
             .enable_save(SessionHistory::new())
-            .enable_reset(Context::new_without_importer)
+            .enable_reset()
             .enable_quit()
     }
 
@@ -580,13 +745,13 @@ mod test {
         input: &'static str,
         expected: BareCommand,
     ) {
-        let cmd = runner.get_command(input, ctx).unwrap().unwrap();
-        assert_eq!(expected, cmd.into_bare());
+        let cmd = runner.parse_command(input, ctx).unwrap().unwrap();
+        assert_eq!(expected, cmd.into());
     }
 
     #[track_caller]
     fn expect_fail(runner: &CommandRunner<()>, ctx: &mut Context, input: &'static str) {
-        assert!(runner.get_command(input, ctx).is_err());
+        assert!(runner.parse_command(input, ctx).is_err());
     }
 
     #[test]
@@ -754,7 +919,22 @@ mod test {
         let mut ctx = Context::new_without_importer();
         let runner = new_runner();
 
-        expect_ok(&runner, &mut ctx, "help", BareCommand::Help);
+        expect_ok(
+            &runner,
+            &mut ctx,
+            "help",
+            BareCommand::Help {
+                help_kind: HelpKind::BasicHelp,
+            },
+        );
+        expect_ok(
+            &runner,
+            &mut ctx,
+            "help commands",
+            BareCommand::Help {
+                help_kind: HelpKind::AllCommands,
+            },
+        );
         expect_fail(&runner, &mut ctx, "help arg");
         expect_fail(&runner, &mut ctx, "help arg1 arg2");
 
@@ -837,7 +1017,7 @@ mod test {
     #[test]
     fn test_runner() {
         fn test_case(
-            runner: &CommandRunner<()>,
+            runner: &mut CommandRunner<()>,
             ctx: &mut Context,
             input: &'static str,
             expected: CommandControlFlow,
@@ -851,25 +1031,30 @@ mod test {
 
         let mut ctx = Context::new_without_importer();
 
-        let runner = CommandRunner::new().print_with(|_| {}).enable_quit();
+        let mut runner = CommandRunner::new().print_with(|_| {}).enable_quit();
 
-        test_case(&runner, &mut ctx, "help", CommandControlFlow::Continue);
-        test_case(&runner, &mut ctx, "list", CommandControlFlow::Continue);
+        test_case(&mut runner, &mut ctx, "help", CommandControlFlow::Continue);
+        test_case(&mut runner, &mut ctx, "list", CommandControlFlow::Continue);
         test_case(
-            &runner,
+            &mut runner,
             &mut ctx,
             // won't be found, but that's fine
             "info item",
             CommandControlFlow::Continue,
         );
-        test_case(&runner, &mut ctx, "clear", CommandControlFlow::NotACommand);
         test_case(
-            &runner,
+            &mut runner,
+            &mut ctx,
+            "clear",
+            CommandControlFlow::NotACommand,
+        );
+        test_case(
+            &mut runner,
             &mut ctx,
             "save dst",
             CommandControlFlow::NotACommand,
         );
-        test_case(&runner, &mut ctx, "quit", CommandControlFlow::Return);
-        test_case(&runner, &mut ctx, "exit", CommandControlFlow::Return);
+        test_case(&mut runner, &mut ctx, "quit", CommandControlFlow::Return);
+        test_case(&mut runner, &mut ctx, "exit", CommandControlFlow::Return);
     }
 }
