@@ -1,12 +1,13 @@
 use codespan_reporting::diagnostic::LabelStyle;
 
 use crate::{
-    interpreter::RuntimeError,
+    NameResolutionError,
+    interpreter::{RuntimeError, RuntimeErrorKind},
     parser::ParseError,
     pretty_print::PrettyPrint,
-    resolver::ResolverError,
+    resolver::{Resolver, ResolverError},
+    span::Span,
     typechecker::{IncompatibleDimensionsError, TypeCheckError},
-    NameResolutionError,
 };
 
 pub type Diagnostic = codespan_reporting::diagnostic::Diagnostic<usize>;
@@ -17,23 +18,29 @@ pub trait ErrorDiagnostic {
 
 impl ErrorDiagnostic for ParseError {
     fn diagnostics(&self) -> Vec<Diagnostic> {
-        vec![Diagnostic::error()
-            .with_message("while parsing")
-            .with_labels(vec![self
-                .span
-                .diagnostic_label(LabelStyle::Primary)
-                .with_message(self.kind.to_string())])]
+        vec![
+            Diagnostic::error()
+                .with_message("while parsing")
+                .with_labels(vec![
+                    self.span
+                        .diagnostic_label(LabelStyle::Primary)
+                        .with_message(self.kind.to_string()),
+                ]),
+        ]
     }
 }
 
 impl ErrorDiagnostic for ResolverError {
     fn diagnostics(&self) -> Vec<Diagnostic> {
         match self {
-            ResolverError::UnknownModule(span, _) => vec![Diagnostic::error()
-                .with_message("while resolving imports in")
-                .with_labels(vec![span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message("Unknown module")])],
+            ResolverError::UnknownModule(span, _) => vec![
+                Diagnostic::error()
+                    .with_message("while resolving imports in")
+                    .with_labels(vec![
+                        span.diagnostic_label(LabelStyle::Primary)
+                            .with_message("Unknown module"),
+                    ]),
+            ],
             ResolverError::ParseErrors(errors) => {
                 errors.iter().flat_map(|e| e.diagnostics()).collect()
             }
@@ -49,25 +56,30 @@ impl ErrorDiagnostic for NameResolutionError {
                 original_item_type,
                 conflict_span,
                 original_span,
-            } => vec![Diagnostic::error()
-                .with_message("identifier clash in definition")
-                .with_labels(vec![
-                    original_span
-                        .diagnostic_label(LabelStyle::Secondary)
-                        .with_message(if let Some(t) = original_item_type.as_ref() {
-                            format!("Previously defined {t} here")
-                        } else {
-                            "Previously defined here".to_owned()
-                        }),
-                    conflict_span
-                        .diagnostic_label(LabelStyle::Primary)
-                        .with_message("identifier is already in use"),
-                ])],
-            NameResolutionError::ReservedIdentifier(span) => vec![Diagnostic::error()
-                .with_message("reserved identifier may not be used")
-                .with_labels(vec![span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message("reserved identifier")])],
+            } => vec![
+                Diagnostic::error()
+                    .with_message("identifier clash in definition")
+                    .with_labels(vec![
+                        original_span
+                            .diagnostic_label(LabelStyle::Secondary)
+                            .with_message(if let Some(t) = original_item_type.as_ref() {
+                                format!("Previously defined {t} here")
+                            } else {
+                                "Previously defined here".to_owned()
+                            }),
+                        conflict_span
+                            .diagnostic_label(LabelStyle::Primary)
+                            .with_message("identifier is already in use"),
+                    ]),
+            ],
+            NameResolutionError::ReservedIdentifier(span) => vec![
+                Diagnostic::error()
+                    .with_message("reserved identifier may not be used")
+                    .with_labels(vec![
+                        span.diagnostic_label(LabelStyle::Primary)
+                            .with_message("reserved identifier"),
+                    ]),
+            ],
         }
     }
 }
@@ -84,10 +96,11 @@ impl ErrorDiagnostic for TypeCheckError {
                 } else {
                     vec![]
                 };
-                d.with_labels(vec![span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message("unknown identifier")])
-                    .with_notes(notes)
+                d.with_labels(vec![
+                    span.diagnostic_label(LabelStyle::Primary)
+                        .with_message("unknown identifier"),
+                ])
+                .with_notes(notes)
             }
             TypeCheckError::IncompatibleDimensions(IncompatibleDimensionsError {
                 operation,
@@ -126,16 +139,11 @@ impl ErrorDiagnostic for TypeCheckError {
             }
             TypeCheckError::NonScalarExponent(span, type_)
             | TypeCheckError::NonScalarFactorialArgument(span, type_) => d
-                .with_labels(vec![span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message(format!("{type_}"))])
+                .with_labels(vec![
+                    span.diagnostic_label(LabelStyle::Primary)
+                        .with_message(format!("{type_}")),
+                ])
                 .with_notes(vec![inner_error]),
-            TypeCheckError::UnsupportedConstEvalExpression(span, _) => d.with_labels(vec![span
-                .diagnostic_label(LabelStyle::Primary)
-                .with_message(inner_error)]),
-            TypeCheckError::DivisionByZeroInConstEvalExpression(span) => d.with_labels(vec![span
-                .diagnostic_label(LabelStyle::Primary)
-                .with_message(inner_error)]),
             TypeCheckError::RegistryError(re) => match re {
                 crate::registry::RegistryError::EntryExists(_) => d.with_notes(vec![inner_error]),
                 crate::registry::RegistryError::UnknownEntry(name, suggestion) => {
@@ -172,26 +180,28 @@ impl ErrorDiagnostic for TypeCheckError {
                 arity,
                 num_args,
             } => {
-                let mut labels = vec![callable_span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message(format!(
-                        "{what}was called with {num}, but takes {range}",
-                        what = if callable_definition_span.is_some() {
-                            ""
-                        } else {
-                            "procedure or function object "
-                        },
-                        num = if *num_args == 1 {
-                            "one argument".into()
-                        } else {
-                            format!("{num_args} arguments")
-                        },
-                        range = if arity.start() == arity.end() {
-                            format!("{}", arity.start())
-                        } else {
-                            format!("{} to {}", arity.start(), arity.end())
-                        }
-                    ))];
+                let mut labels = vec![
+                    callable_span
+                        .diagnostic_label(LabelStyle::Primary)
+                        .with_message(format!(
+                            "{what}was called with {num}, but takes {range}",
+                            what = if callable_definition_span.is_some() {
+                                ""
+                            } else {
+                                "procedure or function object "
+                            },
+                            num = if *num_args == 1 {
+                                "one argument".into()
+                            } else {
+                                format!("{num_args} arguments")
+                            },
+                            range = if arity.start() == arity.end() {
+                                format!("{}", arity.start())
+                            } else {
+                                format!("{} to {}", arity.start(), arity.end())
+                            }
+                        )),
+                ];
                 if let Some(span) = callable_definition_span {
                     labels.insert(
                         0,
@@ -202,9 +212,6 @@ impl ErrorDiagnostic for TypeCheckError {
 
                 d.with_labels(labels)
             }
-            TypeCheckError::TypeParameterNameClash(span, _) => d.with_labels(vec![span
-                .diagnostic_label(LabelStyle::Primary)
-                .with_message(inner_error)]),
             TypeCheckError::IncompatibleTypesInCondition(
                 if_span,
                 then_type,
@@ -300,10 +307,12 @@ impl ErrorDiagnostic for TypeCheckError {
                     ])
                     .with_notes(vec![inner_error])
                 } else {
-                    d.with_labels(vec![argument_span
-                        .diagnostic_label(LabelStyle::Primary)
-                        .with_message(argument_type.to_string())])
-                        .with_notes(vec![inner_error])
+                    d.with_labels(vec![
+                        argument_span
+                            .diagnostic_label(LabelStyle::Primary)
+                            .with_message(argument_type.to_string()),
+                    ])
+                    .with_notes(vec![inner_error])
                 }
             }
             TypeCheckError::IncompatibleTypesInList(
@@ -322,14 +331,20 @@ impl ErrorDiagnostic for TypeCheckError {
                 ])
                 .with_notes(vec![inner_error]),
             TypeCheckError::NoDimensionlessBaseUnit(span, unit_name) => d
-                .with_labels(vec![span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message(inner_error)])
+                .with_labels(vec![
+                    span.diagnostic_label(LabelStyle::Primary)
+                        .with_message(inner_error),
+                ])
                 .with_notes(vec![
                     format!("Use 'unit {unit_name}' for ad-hoc units."),
                     format!("Use 'unit {unit_name}: Scalar = …' for derived units."),
                 ]),
-            TypeCheckError::ForeignFunctionNeedsTypeAnnotations(span, _)
+            TypeCheckError::UnsupportedConstEvalExpression(span, _)
+            | TypeCheckError::DivisionByZeroInConstEvalExpression(span)
+            | TypeCheckError::TypeParameterNameClash(span, _)
+            | TypeCheckError::ForeignFunctionNeedsTypeAnnotations(span, _)
+            | TypeCheckError::UnknownStruct(span, _)
+            | TypeCheckError::ExponentiationNeedsTypeAnnotation(span)
             | TypeCheckError::UnknownForeignFunction(span, _)
             | TypeCheckError::NonRationalExponent(span)
             | TypeCheckError::OverflowInConstExpr(span)
@@ -338,13 +353,15 @@ impl ErrorDiagnostic for TypeCheckError {
             | TypeCheckError::NoFunctionReferenceToGenericFunction(span)
             | TypeCheckError::OnlyFunctionsAndReferencesCanBeCalled(span)
             | TypeCheckError::DerivedUnitDefinitionMustNotBeGeneric(span)
-            | TypeCheckError::MultipleTypedHoles(span) => d.with_labels(vec![span
-                .diagnostic_label(LabelStyle::Primary)
-                .with_message(inner_error)]),
+            | TypeCheckError::MultipleTypedHoles(span) => d.with_labels(vec![
+                span.diagnostic_label(LabelStyle::Primary)
+                    .with_message(inner_error),
+            ]),
             TypeCheckError::MissingDimension(span, dim) => d
-                .with_labels(vec![span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message(format!("Missing dimension '{dim}'"))])
+                .with_labels(vec![
+                    span.diagnostic_label(LabelStyle::Primary)
+                        .with_message(format!("Missing dimension '{dim}'")),
+                ])
                 .with_notes(vec![format!(
                     "This operation requires the '{dim}' dimension to be defined"
                 )]),
@@ -411,9 +428,6 @@ impl ErrorDiagnostic for TypeCheckError {
                     .diagnostic_label(LabelStyle::Secondary)
                     .with_message("Defined here"),
             ]),
-            TypeCheckError::UnknownStruct(span, _name) => d.with_labels(vec![span
-                .diagnostic_label(LabelStyle::Primary)
-                .with_message(inner_error)]),
             TypeCheckError::UnknownFieldInStructInstantiation(field_span, defn_span, _, _) => d
                 .with_labels(vec![
                     field_span
@@ -459,15 +473,13 @@ impl ErrorDiagnostic for TypeCheckError {
                 ])
             }
             TypeCheckError::MissingDimBound(span) => d
-                .with_labels(vec![span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message(inner_error)])
+                .with_labels(vec![
+                    span.diagnostic_label(LabelStyle::Primary)
+                        .with_message(inner_error),
+                ])
                 .with_notes(vec![
-                    "Consider adding `: Dim` after the type parameter".to_owned()
+                    "Consider adding `: Dim` after the type parameter".to_owned(),
                 ]),
-            TypeCheckError::ExponentiationNeedsTypeAnnotation(span) => d.with_labels(vec![span
-                .diagnostic_label(LabelStyle::Primary)
-                .with_message(inner_error)]),
             TypeCheckError::TypedHoleInStatement(span, type_, statement, matches) => {
                 let mut notes = vec![
                     format!("Found a hole of type '{type_}' in the statement:"),
@@ -479,29 +491,41 @@ impl ErrorDiagnostic for TypeCheckError {
                     notes.push(format!("  {}", matches.join(", ")));
                 }
 
-                d.with_labels(vec![span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message(type_)])
-                    .with_message("Found typed hole")
-                    .with_notes(notes)
+                d.with_labels(vec![
+                    span.diagnostic_label(LabelStyle::Primary)
+                        .with_message(type_),
+                ])
+                .with_message("Found typed hole")
+                .with_notes(notes)
             }
         };
         vec![d]
     }
 }
 
-impl ErrorDiagnostic for RuntimeError {
-    fn diagnostics(&self) -> Vec<Diagnostic> {
-        let inner = format!("{self:#}");
+/// Little helper to quickly implement diagnostic for types that requires the resolver
+/// in order to emit a diagnostic.
+pub struct ResolverDiagnostic<'a, E> {
+    pub resolver: &'a Resolver,
+    pub error: &'a E,
+}
 
-        match self {
-            RuntimeError::AssertFailed(span) => vec![Diagnostic::error()
-                .with_message("assertion failed")
-                .with_labels(vec![span
-                    .diagnostic_label(LabelStyle::Primary)
-                    .with_message("assertion failed")])],
-            RuntimeError::AssertEq2Failed(assert_eq2_error) => {
-                vec![Diagnostic::error()
+impl ErrorDiagnostic for ResolverDiagnostic<'_, RuntimeError> {
+    fn diagnostics(&self) -> Vec<Diagnostic> {
+        let mut diag = Vec::new();
+
+        let inner = format!("{:#}", self.error.kind);
+        match &self.error.kind {
+            RuntimeErrorKind::AssertFailed(span) => diag.push(
+                Diagnostic::error()
+                    .with_message("Assertion failed")
+                    .with_labels(vec![
+                        span.diagnostic_label(LabelStyle::Primary)
+                            .with_message("assertion failed"),
+                    ]),
+            ),
+            RuntimeErrorKind::AssertEq2Failed(assert_eq2_error) => diag.push(
+                Diagnostic::error()
                     .with_message("Assertion failed")
                     .with_labels(vec![
                         assert_eq2_error
@@ -513,28 +537,98 @@ impl ErrorDiagnostic for RuntimeError {
                             .diagnostic_label(LabelStyle::Primary)
                             .with_message(format!("{}", assert_eq2_error.rhs)),
                     ])
-                    .with_notes(vec![inner])]
-            }
-            RuntimeError::AssertEq3Failed(assert_eq3_error) => {
+                    .with_notes(vec![inner]),
+            ),
+            RuntimeErrorKind::AssertEq3Failed(assert_eq3_error) => {
                 let (lhs, rhs) = assert_eq3_error.fmt_comparands();
 
-                vec![Diagnostic::error()
-                    .with_message("Assertion failed")
-                    .with_labels(vec![
-                        assert_eq3_error
-                            .span_lhs
-                            .diagnostic_label(LabelStyle::Secondary)
-                            .with_message(lhs),
-                        assert_eq3_error
-                            .span_rhs
-                            .diagnostic_label(LabelStyle::Primary)
-                            .with_message(rhs),
-                    ])
-                    .with_notes(vec![format!("{self:#}")])]
+                diag.push(
+                    Diagnostic::error()
+                        .with_message("Assertion failed")
+                        .with_labels(vec![
+                            assert_eq3_error
+                                .span_lhs
+                                .diagnostic_label(LabelStyle::Secondary)
+                                .with_message(lhs),
+                            assert_eq3_error
+                                .span_rhs
+                                .diagnostic_label(LabelStyle::Primary)
+                                .with_message(rhs),
+                        ])
+                        .with_notes(vec![inner]),
+                )
             }
-            _ => vec![Diagnostic::error()
-                .with_message("runtime error")
-                .with_notes(vec![inner])],
+            _ => diag.push(
+                Diagnostic::error()
+                    .with_message("runtime error")
+                    .with_notes(vec![inner])
+                    .with_labels_iter(
+                        // We're going to join the three first piece of user code that triggered the error
+                        self.error
+                            .backtrace
+                            .iter()
+                            .filter(|(_, span)| {
+                                if cfg!(debug_assertions) {
+                                    true
+                                } else {
+                                    let file =
+                                        self.resolver.files.get(span.code_source_id).unwrap();
+                                    // Everything that starts by prelude was not written by the user and must be ignored
+                                    !file.name().contains("<builtin>")
+                                }
+                            })
+                            .take(3)
+                            .map(|(_, span)| span.diagnostic_label(LabelStyle::Primary)),
+                    ),
+            ),
         }
+
+        diag.push(
+            Diagnostic::help()
+                .with_message("Backtrace:")
+                .with_notes_iter(self.error.backtrace.iter().enumerate().map(
+                    |(i, (fn_name, span))| {
+                        let file = self.resolver.files.get(span.code_source_id).unwrap();
+                        let file_name = file.name();
+                        let (line, col) = position(*span, file.source());
+
+                        // We want to retrieve a rough summary of the error that fits on a single line.
+                        let error_cause =
+                            &file.source()[span.start.as_usize()..span.end.as_usize()];
+                        // " = {}: " => 5 character + the size of the idx in base 10
+                        let target_width = 120 - 5 - (i + 1).ilog10() as usize;
+                        let last_char_idx = error_cause
+                            .char_indices()
+                            .take(target_width)
+                            .map(|(i, _)| i)
+                            .last()
+                            .unwrap_or_default();
+                        let error_cause = &error_cause[..=last_char_idx].replace('\n', "\\n");
+
+                        format!("{i}: {error_cause}\n\tat {fn_name} - {file_name}:{line}:{col}")
+                    },
+                )),
+        );
+
+        diag
     }
 }
+
+/// Return the position in terms of line number and column number
+/// of the span starting position in the source.
+fn position(span: Span, source: &str) -> (usize, usize) {
+    let mut line = 1;
+    let mut last_line_pos = 0;
+    let start = span.start.as_usize();
+
+    for (i, b) in source.bytes().enumerate() {
+        if b == b'\n' {
+            line += 1;
+            last_line_pos = i;
+        }
+        if i == start {
+            break;
+        }
+    }
+    (line, start - last_line_pos)
+}
