@@ -1,10 +1,10 @@
 use compact_str::{CompactString, ToCompactString};
 
+use crate::BaseRepresentationFactor;
 use crate::arithmetic::{Exponent, Power};
 use crate::ast::{TypeExpression, TypeParameterBound};
 use crate::registry::{BaseRepresentation, Registry, Result};
 use crate::span::Span;
-use crate::BaseRepresentationFactor;
 
 #[derive(Default, Clone)]
 pub struct DimensionRegistry {
@@ -94,6 +94,10 @@ impl DimensionRegistry {
     pub fn contains(&self, dimension_name: &str) -> bool {
         self.registry.contains(dimension_name)
     }
+
+    pub fn is_base_dimension(&self, dimension_name: &str) -> bool {
+        self.registry.is_base_dimension(dimension_name)
+    }
 }
 
 #[test]
