@@ -1,3 +1,4 @@
+use crate::ast::ProcedureKind;
 use crate::span::{ByteIndex, Span};
 
 use std::collections::HashMap;
@@ -6,7 +7,15 @@ use thiserror::Error;
 
 #[derive(Debug, Clone, Error, PartialEq, Eq)]
 pub enum TokenizerErrorKind {
-    #[error("Unexpected character: '{character}'")]
+    #[error("Unexpected character: {}",
+        if *character == '\'' {
+            r#""'""#.to_owned()
+        } else if character.is_ascii() {
+            format!("'{}'", character.escape_default())
+        } else {
+            format!("'{character}' (U+{:0>4X})", *character as u32)
+        }
+    )]
     UnexpectedCharacter { character: char },
 
     #[error("Unexpected character in negative exponent")]
@@ -32,6 +41,18 @@ pub enum TokenizerErrorKind {
 
     #[error("Unexpected '{{' inside string interpolation")]
     UnexpectedCurlyInInterpolation,
+
+    #[error("Unexpected closing of {closing_scope_type:?} scope {}", 
+        if let Some(curr_scope_type) = current_scope {
+            format!("while inside a {:#?} scope", curr_scope_type.scope_type)
+        } else {
+            "while not inside any scope".to_owned()
+        }
+    )]
+    UnexpectedScopeClosing {
+        current_scope: Option<Scope>,
+        closing_scope_type: ScopeType,
+    },
 }
 
 #[derive(Debug, Error, PartialEq, Eq)]
@@ -138,6 +159,7 @@ pub enum TokenKind {
 
     // Other
     Newline,
+    Semicolon,
     Eof,
 }
 
@@ -214,21 +236,19 @@ fn is_identifier_continue(c: char) -> bool {
         && c != '⋅'
 }
 
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub enum ScopeType {
+    Curly,
+    String,
+}
+
 /// When scanning a string interpolation like `"foo = {foo}, and bar = {bar}."`,
 /// the tokenizer needs to keep track of where it currently is, because we allow
 /// for (almost) arbitrary expressions inside the {…} part.
-#[cfg_attr(debug_assertions, derive(Debug))]
-enum InterpolationState {
-    /// We are not inside curly braces.
-    Outside,
-    /// We are currently scanning the inner part of an interpolation.
-    Inside,
-}
-
-impl InterpolationState {
-    fn is_inside(&self) -> bool {
-        matches!(self, InterpolationState::Inside)
-    }
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+pub struct Scope {
+    scope_type: ScopeType,
+    scope_start: ByteIndex,
 }
 
 #[cfg_attr(debug_assertions, derive(Debug))]
@@ -238,10 +258,9 @@ struct Tokenizer {
     token_start: ByteIndex,
     code_source_id: usize,
 
-    // Special fields / state for parsing string interpolations
-    string_start: ByteIndex,
-    interpolation_start: ByteIndex,
-    interpolation_state: InterpolationState,
+    // Special state for parsing nested strings and/or structs in string interpolations
+    scopes: Vec<Scope>,
+    last_token: Option<TokenKind>,
 }
 
 fn char_at(s: &str, byte_index: usize) -> Option<char> {
@@ -256,9 +275,8 @@ impl Tokenizer {
             token_start: ByteIndex(0),
 
             code_source_id,
-            string_start: ByteIndex(0),
-            interpolation_start: ByteIndex(0),
-            interpolation_state: InterpolationState::Outside,
+            scopes: Vec::new(),
+            last_token: None,
         }
     }
 
@@ -267,6 +285,7 @@ impl Tokenizer {
         while !self.at_end(input) {
             self.token_start = self.current;
             if let Some(token) = self.scan_single_token(input)? {
+                self.last_token = Some(token.kind.to_owned());
                 tokens.push(token);
             }
         }
@@ -363,9 +382,17 @@ impl Tokenizer {
                     break;
                 }
                 Some('\\') if !escaped => true,
-                Some('"') | Some('{') if !escaped => {
+                Some('"') if !escaped => {
                     break;
                 }
+                c @ (Some('{') | Some('}')) if c != self.peek2(input) => {
+                    break;
+                }
+                Some('{') | Some('}') => {
+                    // extra advance to skip both curly's making up the escape sequence
+                    self.advance(input);
+                    false
+                }
                 Some(_) => false,
             };
 
@@ -375,6 +402,62 @@ impl Tokenizer {
         Ok(())
     }
 
+    fn open_scope(&mut self, scope_type: ScopeType) -> Result<()> {
+        let new_scope = Scope {
+            scope_type,
+            scope_start: self.last,
+        };
+        self.scopes.push(new_scope);
+
+        Ok(())
+    }
+
+    fn close_scope(&mut self, scope_type: ScopeType) -> Result<Scope> {
+        if self.is_directly_inside(scope_type) {
+            let scope = self.scopes.pop().unwrap();
+            Ok(scope)
+        } else {
+            Err(TokenizerError {
+                kind: TokenizerErrorKind::UnexpectedScopeClosing {
+                    current_scope: self.scopes.last().copied(),
+                    closing_scope_type: scope_type,
+                },
+                span: Span {
+                    start: self.last,
+                    end: self.current,
+                    code_source_id: self.code_source_id,
+                },
+            })
+        }
+    }
+
+    fn is_directly_inside(&self, scope_type: ScopeType) -> bool {
+        self.scopes
+            .last()
+            .is_some_and(|scope| scope.scope_type == scope_type)
+    }
+
+    fn is_inside_child_of(&self, scope_type: ScopeType) -> bool {
+        let Some(i) = self.scopes.len().checked_sub(2) else {
+            return false;
+        };
+        self.scopes
+            .get(i)
+            .is_some_and(|scope| scope.scope_type == scope_type)
+    }
+
+    fn scope_start(&self, scope_type: ScopeType) -> Option<ByteIndex> {
+        self.scopes
+            .iter()
+            .filter(|scope| scope.scope_type == scope_type)
+            .next_back()
+            .map(|scope| scope.scope_start)
+    }
+
+    fn is_inside_interpolation(&self) -> bool {
+        self.is_directly_inside(ScopeType::Curly) && self.is_inside_child_of(ScopeType::String)
+    }
+
     fn scan_single_token<'a>(&mut self, input: &'a str) -> Result<Option<Token<'a>>> {
         fn is_ascii_hex_digit(c: char) -> bool {
             c.is_ascii_hexdigit()
@@ -415,10 +498,10 @@ impl Tokenizer {
             m.insert("inf", TokenKind::Inf);
 
             // procedures
-            m.insert("print", TokenKind::ProcedurePrint);
-            m.insert("assert", TokenKind::ProcedureAssert);
-            m.insert("assert_eq", TokenKind::ProcedureAssertEq);
-            m.insert("type", TokenKind::ProcedureType);
+            m.insert(ProcedureKind::Print.name(), TokenKind::ProcedurePrint);
+            m.insert(ProcedureKind::Assert.name(), TokenKind::ProcedureAssert);
+            m.insert(ProcedureKind::AssertEq.name(), TokenKind::ProcedureAssertEq);
+            m.insert(ProcedureKind::Type.name(), TokenKind::ProcedureType);
 
             // type names
             m.insert("Bool", TokenKind::Bool);
@@ -459,8 +542,14 @@ impl Tokenizer {
             ')' => TokenKind::RightParen,
             '[' => TokenKind::LeftBracket,
             ']' => TokenKind::RightBracket,
-            '{' if !self.interpolation_state.is_inside() => TokenKind::LeftCurly,
-            '}' if !self.interpolation_state.is_inside() => TokenKind::RightCurly,
+            '{' if self.last_token == Some(TokenKind::Identifier) => {
+                self.open_scope(ScopeType::Curly)?;
+                TokenKind::LeftCurly
+            }
+            '}' if !self.is_inside_interpolation() => {
+                self.close_scope(ScopeType::Curly)?;
+                TokenKind::RightCurly
+            }
             '≤' => TokenKind::LessOrEqual,
             '<' if self.match_char(input, '=') => TokenKind::LessOrEqual,
             '<' => TokenKind::LessThan,
@@ -540,7 +629,7 @@ impl Tokenizer {
 
                 TokenKind::Ellipsis
             }
-            '.' if self.peek(input).map_or(false, is_identifier_start) => TokenKind::Period,
+            '.' if self.peek(input).is_some_and(is_identifier_start) => TokenKind::Period,
             '.' => {
                 self.consume_stream_of_digits(input, true, true, true)?;
                 self.scientific_notation(input)?;
@@ -551,6 +640,7 @@ impl Tokenizer {
                 return Ok(None);
             }
             '\n' => TokenKind::Newline,
+            ';' => TokenKind::Semicolon,
             '&' if self.match_char(input, '&') => TokenKind::LogicalAnd,
             '|' if self.match_char(input, '|') => TokenKind::LogicalOr,
             '|' if self.match_char(input, '>') => TokenKind::PostfixApply,
@@ -586,41 +676,60 @@ impl Tokenizer {
             '¹' | '²' | '³' | '⁴' | '⁵' | '⁶' | '⁷' | '⁸' | '⁹' => {
                 TokenKind::UnicodeExponent
             }
-            '"' => match self.interpolation_state {
-                InterpolationState::Outside => {
-                    self.string_start = self.token_start;
+            '"' if self.is_inside_interpolation()
+                && matches!(
+                    self.last_token,
+                    Some(TokenKind::StringFixed)
+                        | Some(TokenKind::StringInterpolationEnd)
+                        | Some(TokenKind::Identifier)
+                ) =>
+            {
+                return Err(TokenizerError {
+                    kind: TokenizerErrorKind::UnterminatedStringInterpolation,
+                    span: Span {
+                        start: self
+                            .scope_start(ScopeType::Curly)
+                            .unwrap_or(self.token_start),
+                        end: self.last,
+                        code_source_id: self.code_source_id,
+                    },
+                });
+            }
+            '"' => {
+                self.open_scope(ScopeType::String)?;
+                self.consume_string(input)?;
 
-                    self.consume_string(input)?;
-
-                    if self.match_char(input, '"') {
-                        TokenKind::StringFixed
-                    } else if self.match_char(input, '{') {
-                        self.interpolation_state = InterpolationState::Inside;
-                        self.interpolation_start = self.last;
-                        TokenKind::StringInterpolationStart
-                    } else {
-                        return Err(TokenizerError {
-                            kind: TokenizerErrorKind::UnterminatedString,
-                            span: Span {
-                                start: self.token_start,
-                                end: self.current,
-                                code_source_id: self.code_source_id,
-                            },
-                        });
-                    }
-                }
-                InterpolationState::Inside => {
+                if self.match_char(input, '"') {
+                    self.close_scope(ScopeType::String)?;
+                    TokenKind::StringFixed
+                } else if self.match_char(input, '{') {
+                    self.open_scope(ScopeType::Curly)?;
+                    TokenKind::StringInterpolationStart
+                } else if self.is_inside_interpolation() {
                     return Err(TokenizerError {
                         kind: TokenizerErrorKind::UnterminatedStringInterpolation,
                         span: Span {
-                            start: self.interpolation_start,
-                            end: self.last,
+                            start: self
+                                .scope_start(ScopeType::Curly)
+                                .unwrap_or(self.token_start),
+                            end: self.current,
+                            code_source_id: self.code_source_id,
+                        },
+                    });
+                } else {
+                    return Err(TokenizerError {
+                        kind: TokenizerErrorKind::UnterminatedString,
+                        span: Span {
+                            start: self
+                                .scope_start(ScopeType::String)
+                                .unwrap_or(self.token_start),
+                            end: self.current,
                             code_source_id: self.code_source_id,
                         },
                     });
                 }
-            },
-            ':' if self.interpolation_state.is_inside() => {
+            }
+            ':' if self.is_inside_interpolation() => {
                 while self
                     .peek(input)
                     .map(|c| c != '"' && c != '}')
@@ -652,30 +761,37 @@ impl Tokenizer {
                     });
                 }
             }
-            '}' if self.interpolation_state.is_inside() => {
+            '}' if self.is_inside_interpolation() => {
+                self.close_scope(ScopeType::Curly)?;
                 self.consume_string(input)?;
 
                 if self.match_char(input, '"') {
-                    self.interpolation_state = InterpolationState::Outside;
+                    self.close_scope(ScopeType::String)?;
                     TokenKind::StringInterpolationEnd
                 } else if self.match_char(input, '{') {
-                    self.interpolation_start = self.last;
+                    self.open_scope(ScopeType::Curly)?;
                     TokenKind::StringInterpolationMiddle
                 } else {
                     return Err(TokenizerError {
                         kind: TokenizerErrorKind::UnterminatedString,
                         span: Span {
-                            start: self.string_start,
+                            start: self
+                                .scope_start(ScopeType::String)
+                                .unwrap_or(self.token_start),
                             end: self.current,
                             code_source_id: self.code_source_id,
                         },
                     });
                 }
             }
-            '{' if self.interpolation_state.is_inside() => {
+            '{' if self.is_inside_interpolation() => {
                 return Err(TokenizerError {
                     kind: TokenizerErrorKind::UnexpectedCurlyInInterpolation,
-                    span: self.last.single_character_span(code_source_id),
+                    span: Span {
+                        start: self.last,
+                        end: self.current,
+                        code_source_id: self.code_source_id,
+                    },
                 });
             }
             '…' => TokenKind::Ellipsis,
@@ -765,7 +881,7 @@ impl Tokenizer {
     }
 }
 
-pub fn tokenize(input: &str, code_source_id: usize) -> Result<Vec<Token>> {
+pub fn tokenize(input: &str, code_source_id: usize) -> Result<Vec<Token<'_>>> {
     let mut tokenizer = Tokenizer::new(code_source_id);
     tokenizer.scan(input)
 }
@@ -869,6 +985,16 @@ fn test_tokenize_basic() {
         ]
     );
 
+    assert_eq!(
+        tokenize_reduced("1;42").unwrap(),
+        [
+            ("1", Number, ByteIndex(0)),
+            (";", Semicolon, ByteIndex(1)),
+            ("42", Number, ByteIndex(2)),
+            ("", Eof, ByteIndex(4))
+        ]
+    );
+
     assert_eq!(
         tokenize_reduced("…").unwrap(),
         [
@@ -1112,6 +1238,30 @@ fn test_tokenize_string() {
         ]
     );
 
+    assert_eq!(
+        tokenize_reduced("\"foo = {\"foo\"}, and bar = {\"bar\"}\"").unwrap(),
+        [
+            ("\"foo = {", StringInterpolationStart, ByteIndex(0)),
+            ("\"foo\"", StringFixed, ByteIndex(8)),
+            ("}, and bar = {", StringInterpolationMiddle, ByteIndex(13)),
+            ("\"bar\"", StringFixed, ByteIndex(27)),
+            ("}\"", StringInterpolationEnd, ByteIndex(32)),
+            ("", Eof, ByteIndex(34))
+        ]
+    );
+
+    assert_eq!(
+        tokenize_reduced("\"foo = {\"foo, and bar = {\"bar\"}\"}\"").unwrap(),
+        [
+            ("\"foo = {", StringInterpolationStart, ByteIndex(0)),
+            ("\"foo, and bar = {", StringInterpolationStart, ByteIndex(8)),
+            ("\"bar\"", StringFixed, ByteIndex(25)),
+            ("}\"", StringInterpolationEnd, ByteIndex(30)),
+            ("}\"", StringInterpolationEnd, ByteIndex(32)),
+            ("", Eof, ByteIndex(34))
+        ]
+    );
+
     assert_eq!(
         tokenize("\"foo", 0).unwrap_err().kind,
         TokenizerErrorKind::UnterminatedString
@@ -1120,13 +1270,29 @@ fn test_tokenize_string() {
         tokenize("\"foo = {foo\"", 0).unwrap_err().kind,
         TokenizerErrorKind::UnterminatedStringInterpolation
     );
+    assert_eq!(
+        tokenize("\"foobar = {\"foo{\"bar\"}\"\"", 0)
+            .unwrap_err()
+            .kind,
+        TokenizerErrorKind::UnterminatedStringInterpolation
+    );
     assert_eq!(
         tokenize("\"foo = {foo}.", 0).unwrap_err().kind,
         TokenizerErrorKind::UnterminatedString
     );
     assert_eq!(
-        tokenize("\"foo = {foo, bar = {bar}\"", 0).unwrap_err().kind,
-        TokenizerErrorKind::UnexpectedCurlyInInterpolation
+        tokenize("\"foo = {\"foo\"}.", 0).unwrap_err().kind,
+        TokenizerErrorKind::UnterminatedString
+    );
+    assert_eq!(
+        tokenize("\"foo = {\"foo}.\"", 0).unwrap_err().kind,
+        TokenizerErrorKind::UnterminatedString
+    );
+    assert_eq!(
+        tokenize("\"foobar = {\"foo{\"bar}\"}.\"", 0)
+            .unwrap_err()
+            .kind,
+        TokenizerErrorKind::UnterminatedString
     );
 
     insta::assert_snapshot!(
@@ -1138,9 +1304,9 @@ fn test_tokenize_string() {
     );
 
     insta::assert_snapshot!(
-        tokenize_reduced_pretty(r#""start \{inner\} end""#).unwrap(),
+        tokenize_reduced_pretty(r#""start {{inner}} end""#).unwrap(),
         @r###"
-    "\"start \\{inner\\} end\"", StringFixed, 0
+    "\"start {{inner}} end\"", StringFixed, 0
     "", Eof, 21
     "###
     );
