@@ -1,10 +1,12 @@
 use std::collections::{HashMap, VecDeque};
+use std::fmt::Display;
 use std::sync::Arc;
-use std::{cmp::Ordering, fmt::Display};
 
+use compact_str::{CompactString, ToCompactString};
 use indexmap::IndexMap;
 use num_traits::ToPrimitive;
 
+use crate::interpreter::RuntimeErrorKind;
 use crate::list::NumbatList;
 use crate::span::Span;
 use crate::typed_ast::StructInfo;
@@ -210,9 +212,9 @@ pub enum Constant {
     Scalar(f64),
     Unit(Unit),
     Boolean(bool),
-    String(String),
+    String(CompactString),
     FunctionReference(FunctionReference),
-    FormatSpecifiers(Option<String>),
+    FormatSpecifiers(Option<CompactString>),
 }
 
 impl Constant {
@@ -271,9 +273,9 @@ pub struct ExecutionContext<'a> {
 
 #[derive(Clone)]
 pub struct Vm {
-    /// The actual code of the program, structured by function name. The code
+    /// The actual code and spans of the program, structured by function name. The code
     /// for the global scope is at index 0 under the function name `<main>`.
-    bytecode: Vec<(String, Vec<u8>)>,
+    bytecode: Vec<(CompactString, Vec<u8>, Vec<Span>)>,
 
     /// An index into the `bytecode` vector referring to the function which is
     /// currently being compiled.
@@ -283,7 +285,7 @@ pub struct Vm {
     pub constants: Vec<Constant>,
 
     /// struct metadata, used so we can display struct fields at runtime
-    struct_infos: IndexMap<String, Arc<StructInfo>>,
+    struct_infos: IndexMap<CompactString, Arc<StructInfo>>,
 
     /// Unit prefixes in use
     prefixes: Vec<Prefix>,
@@ -293,15 +295,14 @@ pub struct Vm {
 
     /// Meta information about derived units:
     /// - Unit name
-    /// - Canonical name
     /// - Metadata
-    unit_information: Vec<(String, Option<String>, UnitMetadata)>,
+    unit_information: Vec<(CompactString, UnitMetadata)>,
 
     /// Result of the last expression
     last_result: Option<Value>,
 
     /// List of registered native/foreign functions
-    ffi_callables: Vec<&'static ForeignFunction>,
+    ffi_callables: IndexMap<&'static str, &'static ForeignFunction>,
 
     /// Spans for arguments of procedure calls. This is used for
     /// assertion error messages, for example.
@@ -322,7 +323,7 @@ pub struct Vm {
 impl Vm {
     pub fn new() -> Self {
         Self {
-            bytecode: vec![("<main>".into(), vec![])],
+            bytecode: vec![("<main>".into(), vec![], vec![])],
             current_chunk_index: 0,
             constants: vec![],
             struct_infos: IndexMap::new(),
@@ -330,7 +331,10 @@ impl Vm {
             strings: vec![],
             unit_information: vec![],
             last_result: None,
-            ffi_callables: ffi::procedures().iter().map(|(_, ff)| ff).collect(),
+            ffi_callables: ffi::procedures()
+                .iter()
+                .map(|(kind, ff)| (kind.name(), ff))
+                .collect(),
             procedure_arg_spans: vec![],
             frames: vec![CallFrame::root()],
             stack: vec![],
@@ -342,10 +346,33 @@ impl Vm {
         self.debug = activate;
     }
 
+    pub(crate) fn runtime_error(&self, kind: RuntimeErrorKind) -> RuntimeError {
+        RuntimeError {
+            kind,
+            backtrace: self.backtrace(),
+        }
+    }
+
+    /// Return a list of function name + the span which triggered the error.
+    /// The deepest elements are returned first.
+    pub fn backtrace(&self) -> Vec<(CompactString, Span)> {
+        self.frames
+            .iter()
+            .rev()
+            .map(|cf| {
+                (
+                    self.bytecode[cf.function_idx].0.clone(),
+                    self.bytecode[cf.function_idx].2[cf.ip.saturating_sub(1)],
+                )
+            })
+            .collect()
+    }
+
     // The following functions are helpers for the compilation process
 
-    fn current_chunk_mut(&mut self) -> &mut Vec<u8> {
-        &mut self.bytecode[self.current_chunk_index].1
+    fn current_chunk_mut(&mut self) -> (&mut Vec<u8>, &mut Vec<Span>) {
+        let current = &mut self.bytecode[self.current_chunk_index];
+        (&mut current.1, &mut current.2)
     }
 
     fn push_u16(chunk: &mut Vec<u8>, data: u16) {
@@ -354,29 +381,34 @@ impl Vm {
         chunk.push(arg_bytes[1]);
     }
 
-    pub fn add_op(&mut self, op: Op) {
-        self.current_chunk_mut().push(op as u8);
+    pub fn add_op(&mut self, op: Op, span: Span) {
+        let (bytecode, spans) = self.current_chunk_mut();
+        bytecode.push(op as u8);
+        spans.push(span);
     }
 
-    pub fn add_op1(&mut self, op: Op, arg: u16) {
-        let current_chunk = self.current_chunk_mut();
-        current_chunk.push(op as u8);
-        Self::push_u16(current_chunk, arg)
+    pub fn add_op1(&mut self, op: Op, arg: u16, span: Span) {
+        let (bytecode, spans) = self.current_chunk_mut();
+        bytecode.push(op as u8);
+        Self::push_u16(bytecode, arg);
+        spans.extend(std::iter::repeat_n(span, 3));
     }
 
-    pub(crate) fn add_op2(&mut self, op: Op, arg1: u16, arg2: u16) {
-        let current_chunk = self.current_chunk_mut();
-        current_chunk.push(op as u8);
-        Self::push_u16(current_chunk, arg1);
-        Self::push_u16(current_chunk, arg2);
+    pub(crate) fn add_op2(&mut self, op: Op, arg1: u16, arg2: u16, span: Span) {
+        let (bytecode, spans) = self.current_chunk_mut();
+        bytecode.push(op as u8);
+        Self::push_u16(bytecode, arg1);
+        Self::push_u16(bytecode, arg2);
+        spans.extend(std::iter::repeat_n(span, 5));
     }
 
-    pub(crate) fn add_op3(&mut self, op: Op, arg1: u16, arg2: u16, arg3: u16) {
-        let current_chunk = self.current_chunk_mut();
-        current_chunk.push(op as u8);
-        Self::push_u16(current_chunk, arg1);
-        Self::push_u16(current_chunk, arg2);
-        Self::push_u16(current_chunk, arg3);
+    pub(crate) fn add_op3(&mut self, op: Op, arg1: u16, arg2: u16, arg3: u16, span: Span) {
+        let (bytecode, spans) = self.current_chunk_mut();
+        bytecode.push(op as u8);
+        Self::push_u16(bytecode, arg1);
+        Self::push_u16(bytecode, arg2);
+        Self::push_u16(bytecode, arg3);
+        spans.extend(std::iter::repeat_n(span, 7));
     }
 
     pub fn current_offset(&self) -> u16 {
@@ -385,9 +417,9 @@ impl Vm {
 
     pub fn patch_u16_value_at(&mut self, offset: u16, arg: u16) {
         let offset = offset as usize;
-        let chunk = self.current_chunk_mut();
-        chunk[offset] = (arg & 0xff) as u8;
-        chunk[offset + 1] = ((arg >> 8) & 0xff) as u8;
+        let (bytecode, _spans) = self.current_chunk_mut();
+        bytecode[offset] = (arg & 0xff) as u8;
+        bytecode[offset + 1] = ((arg >> 8) & 0xff) as u8;
     }
 
     pub fn add_constant(&mut self, constant: Constant) -> u16 {
@@ -418,27 +450,23 @@ impl Vm {
         }
     }
 
-    pub fn add_unit_information(
-        &mut self,
-        unit_name: &str,
-        canonical_unit_name: Option<&str>,
-        metadata: UnitMetadata,
-    ) -> u16 {
-        if let Some(idx) = self.unit_information.iter().position(|i| i.0 == unit_name) {
+    pub fn add_unit_information(&mut self, unit_name: &str, metadata: UnitMetadata) -> u16 {
+        if let Some(idx) = self
+            .unit_information
+            .iter()
+            .position(|(name, _)| name == unit_name)
+        {
             return idx as u16;
         }
 
-        self.unit_information.push((
-            unit_name.to_owned(),
-            canonical_unit_name.map(|s| s.to_owned()),
-            metadata,
-        ));
+        self.unit_information
+            .push((unit_name.to_compact_string(), metadata));
         assert!(self.unit_information.len() <= u16::MAX as usize);
         (self.unit_information.len() - 1) as u16 // TODO: this can overflow, see above
     }
 
     pub(crate) fn begin_function(&mut self, name: &str) {
-        self.bytecode.push((name.into(), vec![]));
+        self.bytecode.push((name.into(), vec![], vec![]));
         self.current_chunk_index = self.bytecode.len() - 1
     }
 
@@ -454,7 +482,7 @@ impl Vm {
             .bytecode
             .iter()
             .rev()
-            .position(|(n, _)| n == name)
+            .position(|(n, _, _)| n == name)
             .unwrap();
         let position = self.bytecode.len() - 1 - rev_position;
         assert!(position <= u16::MAX as usize);
@@ -462,14 +490,14 @@ impl Vm {
     }
 
     pub(crate) fn add_foreign_function(&mut self, name: &str, arity: ArityRange) {
-        let ff = ffi::functions().get(name).unwrap();
+        // `key: &'static str`, whereas `name: &'non_static str`
+        let (key, ff) = ffi::functions().get_key_value(name).unwrap();
         assert!(ff.arity == arity);
-        self.ffi_callables.push(ff);
+        self.ffi_callables.insert(key, ff);
     }
 
     pub(crate) fn get_ffi_callable_idx(&self, name: &str) -> Option<u16> {
-        // TODO: this is a linear search that can certainly be optimized
-        let position = self.ffi_callables.iter().position(|ff| ff.name == name)?;
+        let position = self.ffi_callables.get_index_of(name)?;
         assert!(position <= u16::MAX as usize);
         Some(position as u16)
     }
@@ -494,7 +522,7 @@ impl Vm {
         for (idx, identifier) in self.unit_information.iter().enumerate() {
             eprintln!("  {:04} {}", idx, identifier.0);
         }
-        for (idx, (function_name, bytecode)) in self.bytecode.iter().enumerate() {
+        for (idx, (function_name, bytecode, _spans)) in self.bytecode.iter().enumerate() {
             eprintln!(".CODE {idx} ({function_name})");
             let mut offset = 0;
             while offset < bytecode.len() {
@@ -513,8 +541,8 @@ impl Vm {
 
                 let operands_str = operands
                     .iter()
-                    .map(u16::to_string)
-                    .collect::<Vec<String>>()
+                    .map(u16::to_compact_string)
+                    .collect::<Vec<_>>()
                     .join(" ");
 
                 eprint!(
@@ -650,22 +678,19 @@ impl Vm {
 
                     let conversion_value = self.pop_quantity();
 
-                    let unit_information = &self.unit_information[unit_information_idx as usize];
+                    let (unit_name, metadata) =
+                        &self.unit_information[unit_information_idx as usize];
                     let defining_unit = conversion_value.unit();
 
                     let (base_unit_representation, _) = defining_unit.to_base_unit_representation();
 
                     self.unit_registry
-                        .add_derived_unit(
-                            &unit_information.0,
-                            &base_unit_representation,
-                            unit_information.2.clone(),
-                        )
-                        .map_err(RuntimeError::UnitRegistryError)?;
+                        .add_derived_unit(unit_name, &base_unit_representation, metadata.clone())
+                        .map_err(|e| self.runtime_error(RuntimeErrorKind::UnitRegistryError(e)))?;
 
                     self.constants[constant_idx as usize] = Constant::Unit(Unit::new_derived(
-                        &unit_information.0,
-                        unit_information.2.canonical_name.clone(),
+                        unit_name.clone(),
+                        metadata.canonical_name.clone(),
                         *conversion_value.unsafe_value(),
                         defining_unit.clone(),
                     ));
@@ -694,16 +719,19 @@ impl Vm {
                         Op::Add => &lhs + &rhs,
                         Op::Subtract => &lhs - &rhs,
                         Op::Multiply => Ok(lhs * rhs),
-                        Op::Divide => {
-                            Ok(lhs.checked_div(rhs).ok_or(RuntimeError::DivisionByZero)?)
-                        }
+                        Op::Divide => Ok(lhs
+                            .checked_div(rhs)
+                            .ok_or_else(|| self.runtime_error(RuntimeErrorKind::DivisionByZero))?),
                         Op::Power => lhs.power(rhs),
                         // If the user specifically converted the type of a unit, we should NOT simplify this value
                         // before any operations are applied to it
                         Op::ConvertTo => lhs.convert_to(rhs.unit()).map(Quantity::no_simplify),
                         _ => unreachable!(),
                     };
-                    self.push_quantity(result.map_err(RuntimeError::QuantityError)?);
+                    self.push_quantity(
+                        result
+                            .map_err(|e| self.runtime_error(RuntimeErrorKind::QuantityError(e)))?,
+                    );
                 }
                 op @ (Op::AddToDateTime | Op::SubFromDateTime) => {
                     let rhs = self.pop_quantity();
@@ -715,20 +743,20 @@ impl Vm {
 
                     let seconds_i64 = seconds_f64
                         .to_i64()
-                        .ok_or(RuntimeError::DurationOutOfRange)?;
+                        .ok_or_else(|| self.runtime_error(RuntimeErrorKind::DurationOutOfRange))?;
 
                     let span = jiff::Span::new()
                         .try_seconds(seconds_i64)
-                        .map_err(|_| RuntimeError::DurationOutOfRange)?
+                        .map_err(|_| self.runtime_error(RuntimeErrorKind::DurationOutOfRange))?
                         .nanoseconds((seconds_f64.fract() * 1_000_000_000f64).round() as i64);
 
                     self.push(Value::DateTime(match op {
-                        Op::AddToDateTime => lhs
-                            .checked_add(span)
-                            .map_err(|_| RuntimeError::DateTimeOutOfRange)?,
-                        Op::SubFromDateTime => lhs
-                            .checked_sub(span)
-                            .map_err(|_| RuntimeError::DateTimeOutOfRange)?,
+                        Op::AddToDateTime => lhs.checked_add(span).map_err(|_| {
+                            self.runtime_error(RuntimeErrorKind::DateTimeOutOfRange)
+                        })?,
+                        Op::SubFromDateTime => lhs.checked_sub(span).map_err(|_| {
+                            self.runtime_error(RuntimeErrorKind::DateTimeOutOfRange)
+                        })?,
                         _ => unreachable!(),
                     }));
                 }
@@ -739,10 +767,10 @@ impl Vm {
 
                     let duration = lhs
                         .since(&rhs)
-                        .map_err(|_| RuntimeError::DateTimeOutOfRange)?;
+                        .map_err(|_| self.runtime_error(RuntimeErrorKind::DateTimeOutOfRange))?;
                     let duration = duration
                         .total(jiff::Unit::Second)
-                        .map_err(|_| RuntimeError::DurationOutOfRange)?;
+                        .map_err(|_| self.runtime_error(RuntimeErrorKind::DurationOutOfRange))?;
 
                     let ret = Value::Quantity(Quantity::new(
                         Number::from_f64(duration),
@@ -752,22 +780,31 @@ impl Vm {
                     self.push(ret);
                 }
                 op @ (Op::LessThan | Op::GreaterThan | Op::LessOrEqual | Op::GreatorOrEqual) => {
+                    use crate::quantity::QuantityOrdering;
+                    use std::cmp::Ordering;
+
                     let rhs = self.pop_quantity();
                     let lhs = self.pop_quantity();
 
-                    let result = lhs.partial_cmp(&rhs).ok_or_else(|| {
-                        RuntimeError::QuantityError(QuantityError::IncompatibleUnits(
-                            lhs.unit().clone(),
-                            rhs.unit().clone(),
-                        ))
-                    })?;
-
-                    let result = match op {
-                        Op::LessThan => result == Ordering::Less,
-                        Op::GreaterThan => result == Ordering::Greater,
-                        Op::LessOrEqual => result != Ordering::Greater,
-                        Op::GreatorOrEqual => result != Ordering::Less,
-                        _ => unreachable!(),
+                    let result = match lhs.partial_cmp_preserve_nan(&rhs) {
+                        QuantityOrdering::IncompatibleUnits => {
+                            return Err(Box::new(self.runtime_error(
+                                RuntimeErrorKind::QuantityError(QuantityError::IncompatibleUnits(
+                                    lhs.unit().clone(),
+                                    rhs.unit().clone(),
+                                )),
+                            )));
+                        }
+                        QuantityOrdering::NanOperand => false,
+                        QuantityOrdering::Ok(Ordering::Less) => {
+                            matches!(op, Op::LessThan | Op::LessOrEqual)
+                        }
+                        QuantityOrdering::Ok(Ordering::Equal) => {
+                            matches!(op, Op::LessOrEqual | Op::GreatorOrEqual)
+                        }
+                        QuantityOrdering::Ok(Ordering::Greater) => {
+                            matches!(op, Op::GreaterThan | Op::GreatorOrEqual)
+                        }
                     };
 
                     self.push(Value::Boolean(result));
@@ -809,13 +846,19 @@ impl Vm {
                         .expect("Expected factorial operand to be scalar")
                         .to_f64();
 
+                    let order = self.read_u16();
+
                     if lhs < 0. {
-                        return Err(Box::new(RuntimeError::FactorialOfNegativeNumber));
+                        return Err(Box::new(
+                            self.runtime_error(RuntimeErrorKind::FactorialOfNegativeNumber),
+                        ));
                     } else if lhs.fract() != 0. {
-                        return Err(Box::new(RuntimeError::FactorialOfNonInteger));
+                        return Err(Box::new(
+                            self.runtime_error(RuntimeErrorKind::FactorialOfNonInteger),
+                        ));
                     }
 
-                    self.push_quantity(Quantity::from_scalar(math::factorial(lhs)));
+                    self.push_quantity(Quantity::from_scalar(math::factorial(lhs, order)));
                 }
                 Op::JumpIfFalse => {
                     let offset = self.read_u16() as usize;
@@ -850,8 +893,8 @@ impl Vm {
 
                     match &self.ffi_callables[function_idx].callable {
                         Callable::Function(function) => {
-                            let result = (function)(args);
-                            self.push(result?);
+                            let result = (function)(args).map_err(|e| self.runtime_error(*e))?;
+                            self.push(result);
                         }
                         Callable::Procedure(procedure) => {
                             let span_idx = self.read_u16() as usize;
@@ -862,7 +905,7 @@ impl Vm {
                             match result {
                                 std::ops::ControlFlow::Continue(()) => {}
                                 std::ops::ControlFlow::Break(runtime_error) => {
-                                    return Err(Box::new(runtime_error));
+                                    return Err(Box::new(self.runtime_error(runtime_error)));
                                 }
                             }
                         }
@@ -896,10 +939,13 @@ impl Vm {
 
                             match &self.ffi_callables[function_idx].callable {
                                 Callable::Function(function) => {
-                                    let result = (function)(args);
-                                    self.push(result?);
+                                    let result =
+                                        (function)(args).map_err(|e| self.runtime_error(*e))?;
+                                    self.push(result);
                                 }
-                                Callable::Procedure(..) => unreachable!("Foreign procedures can not be targeted by a function reference"),
+                                Callable::Procedure(..) => unreachable!(
+                                    "Foreign procedures can not be targeted by a function reference"
+                                ),
                             }
                         }
                         FunctionReference::TzConversion(tz_name) => {
@@ -907,8 +953,11 @@ impl Vm {
 
                             let dt = self.pop_datetime();
 
-                            let tz = jiff::tz::TimeZone::get(&tz_name)
-                                .map_err(|_| RuntimeError::UnknownTimezone(tz_name))?;
+                            let tz = jiff::tz::TimeZone::get(&tz_name).map_err(|_| {
+                                self.runtime_error(RuntimeErrorKind::UnknownTimezone(
+                                    tz_name.to_string(),
+                                ))
+                            })?;
 
                             let dt = dt.with_time_zone(tz);
 
@@ -923,22 +972,24 @@ impl Vm {
                 }
                 Op::JoinString => {
                     let num_parts = self.read_u16() as usize;
-                    let mut joined = String::new();
+                    let mut joined = CompactString::with_capacity(num_parts);
                     let to_str = |value| match value {
-                        Value::Quantity(q) => q.full_simplify().to_string(),
-                        Value::Boolean(b) => b.to_string(),
-                        Value::String(s) => s,
+                        Value::Quantity(q) => q.full_simplify().to_compact_string(),
+                        Value::Boolean(b) => b.to_compact_string(),
+                        Value::String(s) => s.to_compact_string(),
                         Value::DateTime(dt) => crate::datetime::to_string(&dt),
-                        Value::FunctionReference(r) => r.to_string(),
-                        s @ Value::StructInstance(..) => s.to_string(),
-                        l @ Value::List(_) => l.to_string(),
+                        Value::FunctionReference(r) => r.to_compact_string(),
+                        s @ Value::StructInstance(..) => s.to_compact_string(),
+                        l @ Value::List(_) => l.to_compact_string(),
                         Value::FormatSpecifiers(_) => unreachable!(),
                     };
 
-                    let map_strfmt_error_to_runtime_error = |err| match err {
-                        strfmt::FmtError::Invalid(s) => RuntimeError::InvalidFormatSpecifiers(s),
+                    let map_strfmt_error_to_runtime_error = |this: &Self, err| match err {
+                        strfmt::FmtError::Invalid(s) => {
+                            this.runtime_error(RuntimeErrorKind::InvalidFormatSpecifiers(s))
+                        }
                         strfmt::FmtError::TypeError(s) => {
-                            RuntimeError::InvalidTypeForFormatSpecifiers(s)
+                            this.runtime_error(RuntimeErrorKind::InvalidTypeForFormatSpecifiers(s))
                         }
                         strfmt::FmtError::KeyError(_) => unreachable!(),
                     };
@@ -950,13 +1001,19 @@ impl Vm {
                                     let q = q.full_simplify();
 
                                     let mut vars = HashMap::new();
-                                    vars.insert("value".to_string(), q.unsafe_value().to_f64());
+                                    vars.insert(
+                                        CompactString::const_new("value"),
+                                        q.unsafe_value().to_f64(),
+                                    );
 
                                     let mut str =
                                         strfmt::strfmt(&format!("{{value{specifiers}}}"), &vars)
-                                            .map_err(map_strfmt_error_to_runtime_error)?;
+                                            .map(CompactString::from)
+                                            .map_err(|e| {
+                                                map_strfmt_error_to_runtime_error(self, e)
+                                            })?;
 
-                                    let unit_str = q.unit().to_string();
+                                    let unit_str = q.unit().to_compact_string();
 
                                     if !unit_str.is_empty() {
                                         str += " ";
@@ -967,10 +1024,11 @@ impl Vm {
                                 }
                                 value => {
                                     let mut vars = HashMap::new();
-                                    vars.insert("value".to_string(), to_str(value));
+                                    vars.insert("value".to_owned(), to_str(value).to_string());
 
                                     strfmt::strfmt(&format!("{{value{specifiers}}}"), &vars)
-                                        .map_err(map_strfmt_error_to_runtime_error)?
+                                        .map(CompactString::from)
+                                        .map_err(|e| map_strfmt_error_to_runtime_error(self, e))?
                                 }
                             },
                             Value::FormatSpecifiers(None) => to_str(self.pop()),
@@ -1084,10 +1142,10 @@ fn vm_basic() {
     vm.add_constant(Constant::Scalar(42.0));
     vm.add_constant(Constant::Scalar(1.0));
 
-    vm.add_op1(Op::LoadConstant, 0);
-    vm.add_op1(Op::LoadConstant, 1);
-    vm.add_op(Op::Add);
-    vm.add_op(Op::Return);
+    vm.add_op1(Op::LoadConstant, 0, Span::dummy());
+    vm.add_op1(Op::LoadConstant, 1, Span::dummy());
+    vm.add_op(Op::Add, Span::dummy());
+    vm.add_op(Op::Return, Span::dummy());
 
     let mut print_fn = |_: &Markup| {};
     let mut ctx = ExecutionContext {
